import datetime
import mock

from django.utils import unittest
from iml_common.lib.date_time import IMLDateTime
from iml_common.lib.date_time import FixedOffset
from iml_common.lib.date_time import LocalOffset


class TestIMLDateTime(unittest.TestCase):
    def test_prase(self):
        for str, date_time in [("2016/10/09T19:20:21+1000",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=0,
                                            tzinfo=FixedOffset(600))),
                               ("2016/10/09T19:20:21.12345+1000",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=123450,
                                            tzinfo=FixedOffset(600))),
                               ("2016-10-09 19:20:21-1000",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=0,
                                            tzinfo=FixedOffset(-600))),
                               ("2016-10-09 19:20:21.12345-1000",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=123450,
                                            tzinfo=FixedOffset(-600))),
                               ("19:20:21-1000 2016-10/09",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=0,
                                            tzinfo=FixedOffset(-600))),
                               ("19:20:21.12345-0000 2016/10-09",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=123450,
                                            tzinfo=FixedOffset(0))),
                               ("        19:20:21.12345        -1000              2016/10-09        ",
                                IMLDateTime(year=2016, month=10, day=9, hour=19, minute=20, second=21, microsecond=123450,
                                            tzinfo=FixedOffset(-600))),
                               ("2016/10-09",
                                IMLDateTime(year=2016, month=10, day=9, tzinfo=FixedOffset(0))),
                               ("19:20:21.12345-0000",
                                IMLDateTime(year=1900, month=1, day=1, hour=19, minute=20, second=21, microsecond=123450,
                                            tzinfo=FixedOffset(0))),
                               ("19:20:21-0500",
                                IMLDateTime(year=1900, month=1, day=1, hour=19, minute=20, second=21, microsecond=0,
                                            tzinfo=FixedOffset(-300)))
                               ]:
            self.assertEqual(IMLDateTime.parse(str), date_time)


class TestFixedOffset(unittest.TestCase):
    def test_properties_negative(self):
        fixed_offset = FixedOffset(-300)

        self.assertEqual(fixed_offset.offset, datetime.timedelta(days=-1, hours=19))
        self.assertEqual(fixed_offset.utcoffset(), datetime.timedelta(days=-1, hours=19))
        self.assertEqual(fixed_offset.tzname(), '<-1 day, 19:00:00>')
        self.assertEqual(fixed_offset.dst(), datetime.timedelta(seconds=0))
        self.assertEqual(str(fixed_offset), '-0500')

    def test_properties_positive(self):
        fixed_offset = FixedOffset(567)

        self.assertEqual(fixed_offset.offset, datetime.timedelta(hours=9, minutes=27))
        self.assertEqual(fixed_offset.utcoffset(), datetime.timedelta(hours=9, minutes=27))
        self.assertEqual(fixed_offset.tzname(), '<9:27:00>')
        self.assertEqual(fixed_offset.dst(), datetime.timedelta(seconds=0))
        self.assertEqual(str(fixed_offset), '0927')


class TestLocalOffset(unittest.TestCase):
    class MockDateTime(datetime.datetime):
        now_time = None
        utc_time = None

        @classmethod
        def now(cls, tz=None):
            return cls.now_time

        @classmethod
        def utcnow(cls):
            return cls.utc_time

    def setUp(self):
        self.MockDateTime.utc_time = datetime.datetime(year=2010, month=4, day=20, hour=20, minute=13, second=2)
        self.MockDateTime.now_time = datetime.datetime(year=2010, month=4, day=20, hour=21, minute=13, second=2)

        mock.patch('iml_common.lib.date_time.datetime', self.MockDateTime).start()

    def test_properties_gmt(self):
        local_offset = LocalOffset()

        self.assertEqual(local_offset.offset, datetime.timedelta(hours=1, minutes=0))
        self.assertEqual(local_offset.utcoffset(), datetime.timedelta(hours=1, minutes=0))
        self.assertEqual(local_offset.tzname(), '<1:00:00>')
        self.assertEqual(local_offset.dst(), datetime.timedelta(seconds=0))
        self.assertEqual(str(local_offset), '0100')

    def test_properties_cst(self):
        self.MockDateTime.now_time = datetime.datetime(year=2010, month=4, day=20, hour=13, minute=13, second=2)

        local_offset = LocalOffset()

        self.assertEqual(local_offset.offset, datetime.timedelta(days=-1, hours=17, minutes=0))
        self.assertEqual(local_offset.utcoffset(), datetime.timedelta(days=-1, hours=17, minutes=0))
        self.assertEqual(local_offset.tzname(), '<-1 day, 17:00:00>')
        self.assertEqual(local_offset.dst(), datetime.timedelta(seconds=0))
        self.assertEqual(str(local_offset), '-0700')

    def test_properties_ist(self):
        self.MockDateTime.now_time = datetime.datetime(year=2010, month=4, day=21, hour=01, minute=43, second=2)

        local_offset = LocalOffset()

        self.assertEqual(local_offset.offset, datetime.timedelta(hours=5, minutes=30))
        self.assertEqual(local_offset.utcoffset(), datetime.timedelta(hours=5, minutes=30))
        self.assertEqual(local_offset.tzname(), '<5:30:00>')
        self.assertEqual(local_offset.dst(), datetime.timedelta(seconds=0))
        self.assertEqual(str(local_offset), '0530')



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import os
import shutil
import tempfile

from iml_common.lib.util import PreserveFileAttributes


class NTPConfig(object):
    """ Class to enable IML to automatically configure NTP """

    DEFAULT_CONFIG_FILE = '/etc/ntp.conf'
    # use only a single marker for replace, comment and insert actions on config file
    MARKER = '# IML_EDITv1'
    PREFIX = 'server'
    IBURST = 'iburst'

    def __init__(self, config_file=DEFAULT_CONFIG_FILE, logger=None):
        """
        Init configures instance variables

        :param config_file: destination directory for resultant config file
        :param logger: external logger to write messages to
        """
        self.logger = logger
        self.config_path = config_file
        self.lines = []       # lines to be read in from config file

    def get_configured_server(self, markers):
        """
        Return the IML-configured server found in the ntp conf file

        if IML did not set the server or if any error occurs return None.
        If IML did set the server and IML-markers are found, return server value

        Additional markers can be added as a list parameter, these will then be used to
        determine if the config file has IML edits

        :return: IML-set server found in config, None if no IML-configured server or IOError
        """
        if not markers:
            markers = []
        assert type(markers) is list
        markers.append(self.MARKER)

        # read in file content
        try:
            with open(self.config_path, 'r') as f:
                lines = f.readlines()
        except IOError as e:
            self._log(e.message, level='error')
            return None

        found = None
        # verify this config file has IML custom edits
        for line in lines:
            try:
                found = next(mark for mark in markers if line.find(mark) >= 0)
                # we have found an IML edit in configurations file
                break
            except StopIteration:
                pass
        if not found:
            # no IML edits found
            return None

        # find first server directive in IML-modified file
        try:
            server = next(line.split()[1] for line in lines if line.startswith('server '))
        except StopIteration:
            self._log('no configured server found in IML-modified ntp config file', level='debug')
            return None
        if server == '127.127.1.0':
            # ntp configured to use local clock
            return 'localhost'
        return server

    def _log(self, msg, level='info'):
        """ utility function for using reference to an external logger """
        if self.logger and hasattr(self.logger, level):
            getattr(self.logger, level)('{0}: {1}'.format(self.__class__.__name__, msg))

    def _reset_and_read_conf(self):
        """
        Read the contents of the config file into list of strings/lines, because we are using a
        pre-configured file there should be no IML changes and only standard ntp (v4) formatting
        """
        # read in file content
        with open(self.config_path, 'r') as f:
            original_lines = f.readlines()

        self.lines = []

        for line in original_lines:
            if self.MARKER in line:
                # retrieve commented out directive from line
                line = line[line.find(self.MARKER) + len(self.MARKER) + 1:]
                # don't append edited lines not replacing an old directive
                if line.strip() != "":
                    self.lines.append(line)
            else:
                self.lines.append(line)

    def _write_conf(self):
        """
        Create and open temporary file and write new config content, rename to config
        destination file. Close any open files and file descriptors
        """
        with PreserveFileAttributes(self.config_path):
            tmp_fd, tmp_name = tempfile.mkstemp(dir='/etc')
            os.write(tmp_fd, ''.join(self.lines))
            os.close(tmp_fd)
            shutil.move(tmp_name, self.config_path)

    def _get_prefix_index(self):
        """ Helper method to return first index of string starting with prefix in list """
        try:
            return next(idx for idx, value in enumerate(self.lines) if
                        value.startswith(self.PREFIX + ' '))
        except StopIteration:
            return None

    def _add(self, server):
        """
        Replace server directive in ntp config file data, new server at the same index as the
        first server directive found.
        Appends directive to file if no active server is found in config file.

        Note: we don't make any assumptions as to the location of directives within the file

        :input: server    replacement server to add in config file directive
        """
        prefix_index = self._get_prefix_index()
        first_prefix_index = None
        line_to_replace = '\n'

        if prefix_index is not None:
            # Mark the index of the first directive we find, this is where we want add the server
            first_prefix_index = prefix_index

            # remove this server directive and comment subsequent ones
            line_to_replace = self.lines.pop(prefix_index)
            prefix_index = self._get_prefix_index()

            # while server directive found in lines, comment out line with marker
            while prefix_index is not None:
                self.lines[prefix_index] = ' '.join([self.MARKER, self.lines[prefix_index]])
                prefix_index = self._get_prefix_index()

        # now we have stored the first server directive and commented out others, add the new content
        if server == 'localhost':
            # add local clock to config
            content = ['server  127.127.1.0 %s\n' % self.MARKER,
                       'fudge   127.127.1.0 stratum 10 %s %s' % (self.MARKER, line_to_replace)]
        else:
            content = [' '.join([self.PREFIX, server, self.IBURST, self.MARKER, line_to_replace])]

        for line in content:
            # if no server directives found in file, append content to the end of the file
            # otherwise, once all server directives have been commented out, add content where the first server was
            if first_prefix_index is None:
                self.lines.append(line)
            else:
                self.lines.insert(first_prefix_index, line)
                # increment indexed to preserve order of content list in config file
                first_prefix_index += 1

    def add(self, server):
        """
        Add server directive to the ntp config file, if no server in config then append given hostname in directive
        at end of file, otherwise, replace existing server directive with that provided

        Comment out any additional server directives

        If server parameter is empty/None, reinstate saved backup file to reset any IML changes

        :return: None on success, error string otherwise
        """
        try:
            self._reset_and_read_conf()

            # proceed if we have no error and server name has been provided
            if server:
                self._add(server)

            self._write_conf()
        except (IOError, OSError) as e:
            self._log(e.message, level='error')

            return e.message

        return None



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import os
import re

from ..lib import shell
from filesystem import FileSystem
from ..blockdevices.blockdevice_linux import BlockDeviceLinux


class FileSystemLdiskfs(FileSystem, BlockDeviceLinux):
    # Lustre 2.x's ldiskfs filesystems appear as ext4, maybe we should translate that
    # in the read from blkid. But listing both is safe.
    _supported_filesystems = ['ldiskfs', 'ext4']

    def __init__(self, fstype, device_path):
        super(FileSystemLdiskfs, self).__init__(fstype, device_path)

        self._modules_initialized = False

    @property
    def label(self):
        self._initialize_modules()

        blkid_output = shell.Shell.try_run(['blkid', '-c/dev/null', '-o', 'value', '-s', 'LABEL', self._device_path])

        return blkid_output.strip()

    @property
    def inode_size(self):
        self._initialize_modules()

        dumpe2fs_output = shell.Shell.try_run(['dumpe2fs', '-h', self._device_path])

        return int(re.search("Inode size:\\s*(\\d+)$", dumpe2fs_output, re.MULTILINE).group(1))

    @property
    def inode_count(self):
        self._initialize_modules()

        dumpe2fs_output = shell.Shell.try_run(["dumpe2fs", "-h", self._device_path])

        return int(re.search("Inode count:\\s*(\\d+)$", dumpe2fs_output, re.MULTILINE).group(1))

    # A curiosity with lustre on ldiskfs is that the umount must be on the 'realpath' not the path that was mkfs'd/mounted
    def umount(self):
        return shell.Shell.try_run(["umount", os.path.realpath(self._device_path)])

    def mkfs(self, target_name, options):
        self._initialize_modules()

        shell.Shell.try_run(['mkfs.lustre'] + options + [self._device_path])

        return {'uuid': self.uuid,
                'filesystem_type': self.filesystem_type,
                'inode_size': self.inode_size,
                'inode_count': self.inode_count}

    def mkfs_options(self, target):
        mkfsoptions = []

        if target.inode_size:
            mkfsoptions.append("-I %s" % target.inode_size)
        if target.bytes_per_inode:
            mkfsoptions.append("-i %s" % target.bytes_per_inode)
        if target.inode_count:
            mkfsoptions.append("-N %s" % target.inode_count)

        return mkfsoptions

    def devices_match(self, device1_path, device2_path, device2_uuid):
        """
        Verifies that the devices referenced in the parameters are the same

        :param device1_path: first device string representation
        :param device2_path: second device string representation
        :param device2_uuid: uuid of second device
        :return: return True if both device identifiers reference the same object
        """
        return os.stat(device1_path).st_ino == os.stat(device2_path).st_ino



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import abc
from iml_common.test.command_capture_testcase import CommandCaptureTestCase


class BaseTestBD(object):
    """
    Encapsulate BaseTestBlockDevice within class to stop test runner from running abstract base test methods
    """
    class BaseTestBlockDevice(CommandCaptureTestCase):
        """
        BlockDevice base test class with abstract test methods to ensure minimum coverage of subclasses
        """

        class_override = None
        __metaclass__ = abc.ABCMeta

        @abc.abstractmethod
        def setUp(self):
            super(BaseTestBD.BaseTestBlockDevice, self).setUp()

            self.blockdevice = None

        @abc.abstractmethod
        def test_initialize_modules(self):
            pass

        @abc.abstractmethod
        def test_filesystem_type_unoccupied(self):
            pass

        @abc.abstractmethod
        def test_filesystem_type_occupied(self):
            pass

        @abc.abstractmethod
        def test_uuid(self):
            pass

        @abc.abstractmethod
        def test_preferred_fstype(self):
            pass

        @abc.abstractmethod
        def test_device_type(self):
            pass

        @abc.abstractmethod
        def test_device_path(self):
            pass

        @abc.abstractmethod
        def test_mgs_targets(self):
            pass

        @abc.abstractmethod
        def test_targets(self):
            pass

        @abc.abstractmethod
        def test_property_values(self):
            pass

        def test_import_success(self):
            self.assertIsNone(self.blockdevice.import_())

        def test_export_success(self):
            self.assertIsNone(self.blockdevice.export())

        def test_export_missing(self):
            self.assertIsNone(self.blockdevice.export())



from iml_common.lib.evaluator import safe_eval
from iml_common.test.iml_unit_testcase import ImlUnitTestCase


class TestEvaluator(ImlUnitTestCase):
    def test_good_evaluation(self):
        var1 = 101242123
        var_2 = 12124.242442
        variable__three = "TestCase"
        integer_1 = 99
        integer_2 = 55
        bool_true = True
        bool_false = False

        properties = {"var1": var1,
                      "var_2": var_2,
                      "variable__three": variable__three,
                      "integer_1": integer_1,
                      "integer_2": integer_2,
                      "bool_true": bool_true,
                      "bool_false": bool_false}

        expressions = ["1 * 2",
                       "var1 == var_2",
                       "var1 != var_2",
                       "var1 - var_2",
                       "(var1*var1)-var_2",
                       "(var1 + var1) * (var_2 * var_2)",
                       "(var1 + var1) / (var_2 * var_2)",
                       "var1 + var1 * var_2 * var_2",
                       "integer_1 % integer_2",
                       "var1 < var_2",
                       "var1 > var_2",
                       "var1 <= var_2",
                       "var1 >= var_2",
                       "var1 <= var1",
                       "var1 >= var1",
                       "bool_true == True",
                       "bool_true == True",
                       "bool_true != True",
                       "bool_false == False",
                       "bool_true == bool_true",
                       "bool_false == bool_true",
                       "bool_false == bool_false",
                       "bool_true == bool_false",
                       "(bool_true == bool_false) or bool_true",
                       "bool_true and bool_false",
                       "bool_true and bool_true",
                       "variable__three == variable__three",
                       '(variable__three == "TestCase") and bool_true',
                       "(variable__three == 'TestCase') and bool_true"]

        for expression in expressions:
            self.assertEqual(eval(expression), safe_eval(expression, properties), expression)

    def test_bad_evaluation(self):
        var1 = 101242123
        variable__three = "TestCase"

        properties = {"var1": var1,
                      "variable__three": variable__three}
        expression_exceptions = [("1 / 0", ZeroDivisionError),
                                 ("rubbish", ValueError),
                                 ('eval("ls .")', TypeError),
                                 ("var1 / (var1 - var1)", ZeroDivisionError),
                                 ("var1 / (var1 - var1", SyntaxError),
                                 ("variable_three == 'TestCase'", ValueError)]

        for expression_exception in expression_exceptions:
            with self.assertRaises(expression_exception[1]):
                safe_eval(expression_exception[0], properties)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import mock
import errno

from iml_common.lib.shell import Shell
from iml_unit_testcase import ImlUnitTestCase


class CommandCaptureCommand(object):
    def __init__(self, args, rc=0, stdout='', stderr='', executions_remaining=99999999):
        self.args = args
        self.rc = rc
        self.stdout = stdout
        self.stderr = stderr
        self.executions_remaining = executions_remaining

    def __str__(self):
        return '"%s" returning %s' % (' '.join(self.args), self.rc)


class CommandCaptureTestCase(ImlUnitTestCase):
    CommandNotFound = 123456

    def setUp(self):
        super(CommandCaptureTestCase, self).setUp()

        self.reset_command_capture()
        self._missing_command_err_msg = 'Command attempted "%s" was unknown to CommandCaptureTestCase, did you intend this?'

        assert 'fake' not in str(Shell.run)
        mock.patch('iml_common.lib.shell.BaseShell.run', self._fake_run).start()

    def _fake_run(self, arg_list, logger=None, monitor_func=None, timeout=Shell.SHELLTIMEOUT, shell=False):
        assert type(arg_list) in [list, str, unicode], 'arg list must be list or str :%s' % type(arg_list)

        # Allow simple commands to just be presented as a string. However do not start formatting the string this
        # will be rejected in a code review. If it has args present them as a list.
        if type(arg_list) in [str, unicode]:
            arg_list = arg_list.split()

        args = tuple(arg_list)
        self._commands_history.append(args)

        try:
            result = self._get_executable_command(args)
            result.executions_remaining -= 1

            if result.rc == self.CommandNotFound:
                raise OSError(errno.ENOENT, result.stderr)

            return Shell.RunResult(result.rc, result.stdout, result.stderr, False)
        except KeyError:
            raise OSError(errno.ENOENT, self._missing_command_err_msg % ' '.join(arg_list))

    def _get_executable_command(self, args):
        '''
        return the command whose args match those given. note that exact order match is needed

        FIXME: return more information about whether command was nearly correct compare each
        command in list, print the one that's closest rather than just 'no such...'

        :param args: Tuple of the arguments of the command
        :return: The command requested or raise a KeyError
        '''
        for command in self._commands:
            if command.args == args and command.executions_remaining > 0:
                return command
        raise KeyError

    def assertRanCommand(self, command):
        '''
        assert that the command made up of the args passed was executed.
        :param command: The command to check was run.
        '''
        self.assertTrue(command.args in self._commands_history)

    def assertRanAllCommandsInOrder(self):
        '''
        assert that all the commands expected were run in the order expected.
        '''
        self.assertEqual(len(self._commands), len(self._commands_history))

        for ran, expected_args in zip(self._commands, self._commands_history):
            self.assertEqual(ran.args, expected_args)

    def assertRanAllCommands(self):
        '''
        assert that all the commands expected were run, the order is not important.
        Some commands may be run more than once.
        '''
        commands_history = set(self._commands_history)

        self.assertEqual(len(self._commands), len(commands_history))

        for command in self._commands:
            self.assertRanCommand(command)

    def add_commands(self, *commands):
        '''
        Add a list of command that is expected to be run.
        :param *commands: CommandCaptureCommand args of commands to be run
        :return: No return value
        '''
        for command in commands:
            self._commands.append(command)

    def single_commands(self, *commands):
        '''
        Add a list of commands that are expected to be run only once.
        :param *commands: CommandCaptureCommand args of commands to be run once
        :return: No return value
        '''
        for command in commands:
            command.executions_remaining = 1
            self._commands.append(command)

    def add_command(self, args, rc = 0, stdout = "", stderr = "", executions_remaining=99999999):
        '''
        Add a single command that is expected to be run.
        :param args: Tuple of the arguments of the command
        :param rc: return of the command
        :param stdout: stdout for the command
        :param stderr: stderr for the command
        :return: No return value
        '''
        self.add_commands(CommandCaptureCommand(args, rc, stdout, stderr, executions_remaining))

    @property
    def commands_ran_count(self):
        return len(self._commands_history)

    def reset_command_capture(self):
        '''
        Reset the command capture to the initialized state.
        :return: None
        '''
        self._commands = []
        self.reset_command_capture_logs()

    def reset_command_capture_logs(self):
        '''
        Reset the command capture logs to the initialized state.
        :return: None
        '''
        self._commands_history = []



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import subprocess
import os
import tempfile
import collections
import time
import warnings


class BaseShell(object):
    class CommandExecutionError(Exception):
        def __init__(self, result, command):
            self.result = result
            self.command = " ".join(command)

        def __str__(self):
            return "Error (%s) running '%s': '%s' '%s'" % (self.result.rc, self.command,
                                                           self.result.stdout, self.result.stderr)

        # Keeping for backwards compatibility, will be removed over time.
        @property
        def rc(self):
            warnings.warn("Direct use of rc in CommandExecutionError is deprecated", DeprecationWarning)
            return self.result.rc

        @property
        def stdout(self):
            warnings.warn("Direct use of stdout in CommandExecutionError is deprecated", DeprecationWarning)
            return self.result.stdout

        @property
        def stderr(self):
            warnings.warn("Direct use of stderr in CommandExecutionError is deprecated", DeprecationWarning)
            return self.result.stderr

    # Effectively no autotimeout for commands.
    SHELLTIMEOUT = 0xfffffff

    RunResult = collections.namedtuple("RunResult", ['rc', 'stdout', 'stderr', 'timeout'])

    @classmethod
    def _run(cls, arg_list, logger, monitor_func, timeout, shell=False):
        """Separate the bare inner of running a command, so that tests can
        stub this function while retaining the related behaviour of run()
        """

        assert type(arg_list) in [list, str, unicode], 'arg list must be list or str :%s' % type(arg_list)

        # Allow simple commands to just be presented as a string. However do not start formatting the string this
        # will be rejected in a code review. If it has args present them as a list.
        if type(arg_list) in [str, unicode]:
            arg_list = arg_list.split()

        # Popen has a limit of 2^16 for the output if you use subprocess.PIPE (as we did recently) so use real files so
        # the output side is effectively limitless
        stdout_fd = tempfile.TemporaryFile()
        stderr_fd = tempfile.TemporaryFile()

        try:
            p = subprocess.Popen(arg_list,
                                 stdout=stdout_fd,
                                 stderr=stderr_fd,
                                 close_fds=True,
                                 shell=shell)

            # Rather than using p.wait(), we do a slightly more involved poll/backoff, in order
            # to poll the thread_state.teardown event as well as the completion of the subprocess.
            # This is done to allow cancellation of subprocesses
            rc = None
            max_wait = 1.0
            wait = 1.0E-3
            timeout += time.time()
            while rc is None:
                rc = p.poll()
                if rc is None:
                    if monitor_func:
                        monitor_func(p, arg_list, logger)

                    time.sleep(1)

                    if time.time() > timeout:
                        p.kill()
                        stdout_fd.seek(0)
                        stderr_fd.seek(0)
                        return cls.RunResult(254,
                                             stdout_fd.read().decode('ascii', 'ignore'),
                                             stderr_fd.read().decode('ascii', 'ignore'),
                                             True)
                    elif wait < max_wait:
                        wait *= 2.0
                else:
                    stdout_fd.seek(0)
                    stderr_fd.seek(0)
                    return cls.RunResult(rc,
                                         stdout_fd.read().decode('ascii', 'ignore'),
                                         stderr_fd.read().decode('ascii', 'ignore'),
                                         False)
        finally:
            stdout_fd.close()
            stderr_fd.close()

    @classmethod
    def run(cls, arg_list, logger=None, monitor_func=None, timeout=SHELLTIMEOUT, shell=False):
        """Run a subprocess, and return a tuple of rc, stdout, stderr.
        Record subprocesses run and their results in log.

        Note: we buffer all output, so do not run subprocesses with large outputs
        using this function.
        """

        # TODO: add a 'quiet' flag and use it from spammy/polling plugins to avoid
        # sending silly amounts of command output back to the manager.
        if logger:
            logger.debug("Shell.run: %s" % repr(arg_list))

        os.environ["TERM"] = ""

        result = cls._run(arg_list, logger, monitor_func, timeout, shell=shell)

        return result

    @classmethod
    def try_run(cls, arg_list, logger=None, monitor_func=None, timeout=SHELLTIMEOUT, shell=False):
        """Run a subprocess, and raise an exception if it returns nonzero.  Return
        stdout string.
        """

        result = cls.run(arg_list, logger, monitor_func, timeout, shell=shell)

        if result.rc != 0:
            raise cls.CommandExecutionError(result, arg_list)

        return result.stdout

    @classmethod
    def run_canned_error_message(cls, arg_list):
        """
        Run a shell command return None is successful, or User Error message if not
        :param args:
        :return: None if successful or canned user error message
        """
        result = cls.run(arg_list)

        if result.rc != 0:
            return "Error (%s) running '%s': '%s' '%s'" % (result.rc, " ".join(arg_list),
                                                           result.stdout, result.stderr)

        return None

# By default Shell is this BaseShell class, and other iml_common modules use BaseShell via Shell by default.
# However consumers (namely the agent today) may change Shell to reference their own SubClass version and this will
# mean that iml_common consumers of Shell use the SubClass rather than the base.
Shell = BaseShell


def set_shell(new_shell_class):
    """
    Change the Shell the iml_common (and any other referencers of Shell in this module) use for Shell commands
    :param new_shell_class: The new Shell classs to use
    """
    global Shell
    Shell = new_shell_class



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


from collections import namedtuple, defaultdict
from ..lib import util, shell
import abc

_cached_device_types = {}


class BlockDevice(object):
    """
    BlockDevice abstraction which provides blockdevice specific functionality
    This class really really really needs to be in a common place to all code
    so that its functionality can be used in all components. Then we could pass
    it around as a class and not as a hash of its values.
    """

    class_override = None
    __metaclass__ = abc.ABCMeta

    # The split for multiple properties in the lustre configuration storage seems to be inconsistent
    # so create a look up table for the splitters.
    lustre_property_delimiters = defaultdict(lambda: "")
    lustre_property_delimiters['failover.node'] = ':'
    lustre_property_delimiters['mgsnode'] = ':'

    class UnknownBlockDevice(KeyError):
        pass

    def __new__(cls, device_type, device):
        try:
            # It is possible the caller doesn't know the device type, but they do know the path - these cases should be
            # avoided but in IML today this is the case, so keep a class variable to allow use to resolve it. We default
            # very badly to linux if we don't have a value.
            if device_type == None:
                if device in _cached_device_types:
                    device_type = _cached_device_types[device]
                else:
                    device_type = 'linux'
            else:
                _cached_device_types[device] = device_type

            subtype = next(klass for klass in util.all_subclasses(BlockDevice) if device_type in klass._supported_device_types)

            if cls != subtype:
                return subtype.__new__(subtype, device_type, device)
            else:
                return super(BlockDevice, cls).__new__(cls)

        except StopIteration:
            raise cls.UnknownBlockDevice("DeviceType %s unknown" % device_type)

    def __init__(self, device_type, device_path):
        self._device_type = device_type
        self._device_path = device_path

    def _initialize_modules(self):
        """
        Called before operations that might result the filesystem specific modules to be loaded.

        :return: None
        """
        return None

    @abc.abstractproperty
    def filesystem_type(self):
        """
        Return type of occupying filesystem(s)
        """
        pass

    @abc.abstractproperty
    def filesystem_info(self):
        """
        Return message regarding occupying filesystem(s) that reside on this block device
        """
        pass

    @abc.abstractproperty
    def uuid(self):
        pass

    @abc.abstractproperty
    def preferred_fstype(self):
        pass

    @property
    def device_type(self):
        return self._device_type

    @property
    def device_path(self):
        return self._device_path

    @classmethod
    def initialise_driver(cls, managed_mode):
        return None

    @classmethod
    def terminate_driver(cls):
        return None

    @abc.abstractmethod
    def mgs_targets(self, log):
        """
        Creates a list of all the mgs targets on a given device, returning a dict of filesystems and names

        :param log: The log to write debug info to
        :return: dict of filesystems and names
        """
        pass

    TargetsInfo = namedtuple('TargetsInfo', ['names', 'params'])

    @abc.abstractmethod
    def targets(self, uuid_name_to_target, device, log):
        pass

    def import_(self):
        """
        If appropriate import the blockdevice, this is required for many devices where only 1 node may have the device
        imported at a time. zpools for example.

        :return: None on success or error message on failure
        """
        return None

    def export(self):
        """
        If appropriate export the blockdevice, this is required for many devices where only 1 node may have the device
        imported at a time. zpools for example.

        :return: None on success or error message on failure
        """
        return None

    def purge_filesystem_configuration(self, filesystem_name, log):
        """
        Purge the details of the filesystem from the mgs blockdevice.  This routine presumes that the blockdevice
        is the mgs_blockdevice and does not make any checks

        :param filesystem_name: The name of the filesystem to purge
        :param log: The logger to use for log messages.
        :return: None on success or error message on failure
        """
        shell_result = shell.Shell.run(['lctl', 'erase_lcfg', filesystem_name])

        if shell_result.rc != 0:
            return "Purge filesystem failed to purge %s with error '%s'" % (filesystem_name, shell_result.stderr)



import os
import shutil
import mock
import time

from errno import ESRCH
from random import randint
from collections import defaultdict
from lockfile import LockFile

from iml_common.test.command_capture_testcase import CommandCaptureCommand, CommandCaptureTestCase
from iml_common.blockdevices.blockdevice_zfs import ZfsDevice
from iml_common.lib.util import ExceptionThrowingThread
from iml_common.lib.shell import BaseShell

TEST_ZPOOL_LOCK_DIR = '/tmp/iml-common-test-locks'
TEST_LOCK_ACQUIRE_TIMEOUT = 1


class TestZfsDevice(CommandCaptureTestCase):
    def setUp(self):
        super(TestZfsDevice, self).setUp()

        self.zpool_name = 'Dave'
        ZfsDevice.lock_refcount = defaultdict(int)

        self.addCleanup(mock.patch.stopall)

    def _add_import_commands(self, name='Boris'):
        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'), stdout=name))

        if name != self.zpool_name:
            self.add_commands(CommandCaptureCommand(('zpool', 'import', '-N', '-o', 'readonly=on', '-o', 'cachefile=none', self.zpool_name)))

    def _add_export_commands(self):
        self.add_commands(CommandCaptureCommand(('zpool', 'export', self.zpool_name)))

    @staticmethod
    def _get_zfs_device(name, try_import):
        device = ZfsDevice(name, try_import)
        device.lock_unique_id = 'mockhost_mockthread.mockpid:%s' % name
        return device


class TestZfsDeviceImportExport(TestZfsDevice):
    def setUp(self):
        super(TestZfsDeviceImportExport, self).setUp()

        mock_lock_file = mock.Mock(autospec=LockFile)
        # add attributes to mock that would be created at runtime and therefore cannot be autospecced
        mock_lock_file.unique_name = ''
        mock_lock_file.pid = 12345

        mock.patch('iml_common.blockdevices.blockdevice_zfs.LockFile',
                   return_value=mock_lock_file).start()
        mock.patch('__builtin__.open').start()

        self.zfs_device = ZfsDevice(self.zpool_name, False)
        self.zfs_device.lock_unique_id = 'mockhost_mockthread.mockpid:%s' % self.zpool_name
        self.zfs_device.lock.path = '%s/%s' % (ZfsDevice.ZPOOL_LOCK_DIR, self.zpool_name)
        self.zfs_device.lock.lock_file = self.zfs_device.lock.path + '.lock'
        self.zfs_device.lock.reset_mock()

        self.zfs_device.lock.acquire.return_value = None
        self.zfs_device.lock.acquire.side_effect = self._toggle_lock
        self.zfs_device.lock.release.return_value = None
        self.zfs_device.lock.release.side_effect = self._toggle_lock
        self.zfs_device.lock.is_locked.return_value = False
        self.zfs_device.lock.i_am_locking.return_value = False

    def _toggle_lock(self, timeout=0):
        self.zfs_device.lock.i_am_locking.return_value ^= True
        self.zfs_device.lock.is_locked.return_value ^= True

    def test_import(self):
        for readonly in [True, False]:
            self.reset_command_capture()

            self.add_commands(CommandCaptureCommand(('zpool', 'import') +
                                                    (('-N', '-o', 'readonly=on', '-o', 'cachefile=none') if readonly else ()) +
                                                    (self.zpool_name,)))

            self.zfs_device.import_(readonly)

            self.zfs_device.lock.acquire.assert_called_once_with(ZfsDevice.LOCK_ACQUIRE_TIMEOUT)
            self.zfs_device.lock.release.assert_called_once_with()
            self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)

            self.assertRanAllCommandsInOrder()
            self.zfs_device.lock.reset_mock()

    def test_import_writable(self):
        self._add_import_commands()
        self._add_export_commands()

        self.assertFalse(self.zfs_device.try_import)
        self.zfs_device.try_import = True

        self.assertTrue(self.zfs_device.__enter__().available)

        self.assertTrue(self.zfs_device.lock.i_am_locking())
        self.assertTrue(self.zfs_device.pool_imported)
        self.assertTrue(self.zfs_device.available)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 1)

        self.zfs_device.__exit__(0, 0, 0)

        self.assertFalse(self.zfs_device.pool_imported)
        self.assertFalse(self.zfs_device.lock.i_am_locking())
        self.assertFalse(self.zfs_device.pool_imported)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)

        self.assertRanAllCommandsInOrder()

    def test_simple_multiple_entry_lock(self):
        exported_zfs_devices = []

        for count in range(1, 10):
            self._add_import_commands()
            exported_zfs_devices.insert(0, self._get_zfs_device(self.zpool_name, True))
            self.assertEqual(exported_zfs_devices[0].__enter__().available, True)
            self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), count)
            self.assertTrue(exported_zfs_devices[0].lock.i_am_locking())
            self.assertTrue(exported_zfs_devices[0].pool_imported)

        for count in range(9, 0, -1):
            self._add_export_commands()
            exported_zfs_devices[0].__exit__(0, 0, 0)
            self.assertFalse(exported_zfs_devices[0].pool_imported)
            self.assertEqual(exported_zfs_devices[0].lock.i_am_locking(),
                             True if ((count - 1) > 0) else False)
            self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), count - 1)
            del exported_zfs_devices[0]

        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)
        self.assertFalse(ZfsDevice(self.zpool_name, True).lock.is_locked())

        self.assertRanAllCommandsInOrder()

    def test_simple_lock_no_import(self):
        self._add_import_commands(self.zpool_name)

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        self.assertEqual(exported_zfs_device.__enter__().available, True)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 1)
        self.assertTrue(exported_zfs_device.lock.i_am_locking())
        self.assertFalse(exported_zfs_device.pool_imported)

        exported_zfs_device.__exit__(0, 0, 0)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)
        self.assertFalse(exported_zfs_device.lock.i_am_locking())
        self.assertFalse(exported_zfs_device.pool_imported)

        self.assertRanAllCommandsInOrder()

    def test_error_in_list(self):
        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'), rc=1))

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        with self.assertRaises(BaseShell.CommandExecutionError):
            exported_zfs_device.__enter__()

        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)
        self.assertFalse(exported_zfs_device.lock.i_am_locking())
        self.assertFalse(exported_zfs_device.pool_imported)
        self.assertFalse(exported_zfs_device.available)

        self.assertRanAllCommandsInOrder()

    def test_error_in_import(self):
        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'), stdout='Boris'),
                          CommandCaptureCommand(('zpool', 'import', '-N', '-o', 'readonly=on', '-o', 'cachefile=none',
                                                 self.zpool_name), rc=1))

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        self.assertFalse(exported_zfs_device.__enter__().available)

        self.assertFalse(exported_zfs_device.pool_imported)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 1)

        exported_zfs_device.__exit__(0, 0, 0)

        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)

        self.assertRanAllCommandsInOrder()

    def test_error_in_export(self):
        self._add_import_commands()
        self.add_commands(CommandCaptureCommand(('zpool', 'export', self.zpool_name), rc=1))

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        self.assertTrue(exported_zfs_device.__enter__().available)

        self.assertTrue(exported_zfs_device.pool_imported)
        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 1)

        with self.assertRaises(BaseShell.CommandExecutionError):
            exported_zfs_device.__exit__(0, 0, 0)

        self.assertEqual(ZfsDevice.lock_refcount.get(self.zfs_device.lock_unique_id), 0)

        self.assertRanAllCommandsInOrder()

    def test_busy_export(self):
        self._add_import_commands()
        self.add_commands(CommandCaptureCommand(('zpool', 'export', self.zpool_name),
                                                rc=1,
                                                stderr="cannot export '%s': pool is busy" % self.zpool_name,
                                                executions_remaining=1))
        self._add_export_commands()

        mock.patch('iml_common.blockdevices.blockdevice_zfs.time.sleep').start()

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        self.assertEqual(exported_zfs_device.__enter__().available, True)
        exported_zfs_device.__exit__(0, 0, 0)

        self.assertRanAllCommandsInOrder()

    def test_very_busy_export(self):
        self._add_import_commands()
        self.add_commands(CommandCaptureCommand(('zpool', 'export', self.zpool_name),
                                                rc=1,
                                                stderr="cannot export '%s': pool is busy" % self.zpool_name))
        self._add_export_commands()

        mock.patch('iml_common.blockdevices.blockdevice_zfs.time.sleep').start()

        exported_zfs_device = self._get_zfs_device(self.zpool_name, True)

        self.assertEqual(exported_zfs_device.__enter__().available, True)

        with self.assertRaises(BaseShell.CommandExecutionError):
            exported_zfs_device.__exit__(0, 0, 0)

    def test_no_import_silent(self):
        exported_zfs_device = self._get_zfs_device(self.zpool_name, False)

        self.assertEqual(exported_zfs_device.available, True)
        self.assertEqual(exported_zfs_device.__enter__().available, True)
        self.assertFalse(exported_zfs_device.pool_imported)
        exported_zfs_device.__exit__(0, 0, 0)

        self.assertRanAllCommandsInOrder()


class TestZfsDeviceLockFile(TestZfsDevice):
    def setUp(self):
        super(TestZfsDeviceLockFile, self).setUp()

        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.ZPOOL_LOCK_DIR',
                   TEST_ZPOOL_LOCK_DIR).start()
        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.LOCK_ACQUIRE_TIMEOUT',
                   TEST_LOCK_ACQUIRE_TIMEOUT).start()

        self.zpool_name = 'Dave'
        self.thread_running = False

        shutil.rmtree(ZfsDevice.ZPOOL_LOCK_DIR, ignore_errors=True)
        os.makedirs(ZfsDevice.ZPOOL_LOCK_DIR)

    def tearDown(self):
        shutil.rmtree(ZfsDevice.ZPOOL_LOCK_DIR, ignore_errors=True)

    def import_(self):
        self.add_command(('zpool', 'import', self.zpool_name))
        self.thread_running = True
        zfs_device = ZfsDevice(self.zpool_name, False)
        zfs_device.import_(False)
        self.assertFalse(zfs_device.lock.i_am_locking())
        self.thread_running = False

    def export(self):
        self._add_export_commands()
        self.thread_running = True
        ZfsDevice(self.zpool_name, False).export()
        self.thread_running = False

    def _blocking_test_base(self, target_func):
        self._add_import_commands()
        self._add_export_commands()

        exported_zfs_device = ZfsDevice(self.zpool_name, True)
        self.assertEqual(exported_zfs_device.lock_unique_id, '%s:%s' % (exported_zfs_device.lock.unique_name,
                                                                        self.zpool_name))

        self.assertTrue(exported_zfs_device.__enter__().available)
        self.assertEqual(ZfsDevice.lock_refcount.get(exported_zfs_device.lock_unique_id), 1)
        self.assertTrue(exported_zfs_device.lock.i_am_locking())
        self.assertTrue(exported_zfs_device.pool_imported)

        # Now this thread should block.
        thread = ExceptionThrowingThread(target=target_func,
                                         args=(),
                                         use_threads=True)
        thread.start()

        time.sleep(0.1)
        self.assertEqual(self.thread_running, True)
        time.sleep(0.1)
        self.assertEqual(self.thread_running, True)

        self.assertEqual(ZfsDevice.lock_refcount.get(exported_zfs_device.lock_unique_id), 1)

        exported_zfs_device.__exit__(0, 0, 0)
        self.assertFalse(exported_zfs_device.lock.i_am_locking())
        self.assertFalse(exported_zfs_device.pool_imported)

        thread.join()

        time.sleep(0.1)
        self.assertEqual(self.thread_running, False)

    def test_import_blocks(self):
        self._blocking_test_base(self.import_)

    def test_export_blocks(self):
        self._blocking_test_base(self.export)

    def test_not_locked(self):
        """ check pid is inserted into lockfile after being acquired """
        zfs_device = ZfsDevice(self.zpool_name, False)

        zfs_device.lock_pool()

        with open(zfs_device.lock.lock_file) as f:
            contents = f.readlines()

        self.assertEqual(zfs_device.lock.pid, os.getpid())
        self.assertEqual(contents, [str(zfs_device.lock.pid)])

    def _create_owned_lockfile(self):
        zfs_device = ZfsDevice(self.zpool_name, False)

        assert zfs_device.lock_unique_id == '%s:%s' % (zfs_device.lock.unique_name, self.zpool_name)
        assert zfs_device.lock.lock_file == '%s/%s.lock' % (ZfsDevice.ZPOOL_LOCK_DIR, self.zpool_name)

        # create lock file and write another processes pid into its content
        other_pid = int(zfs_device.lock.pid)

        while other_pid == int(zfs_device.lock.pid):
            other_pid = randint(1000, 2000)

        with open(zfs_device.lock.lock_file, 'w') as f:
            f.writelines(str(other_pid))

        return zfs_device

    def test_locked_existing_pid(self):
        """
        check lock action blocks if owner pid is running, the fourth time lock.acquire is called, the process givens
        by the PID is found to be non-existent and the lock is broken and reacquired and given the current process' PID
        """
        zfs_device = self._create_owned_lockfile()

        no_such_process_exception = OSError()
        no_such_process_exception.errno = ESRCH
        mock_kill = mock.Mock()
        mock_kill.side_effect = (True, True, True, no_such_process_exception)

        with open(zfs_device.lock.lock_file) as f:
            contents = f.readlines()

        self.assertNotEqual(contents, [str(zfs_device.lock.pid)])

        with mock.patch('os.kill', mock_kill):
            self.assertFalse(zfs_device.lock.i_am_locking())
            zfs_device.lock_pool()

        self.assertTrue(zfs_device.lock.i_am_locking())

        with open(zfs_device.lock.lock_file) as f:
            contents = f.readlines()

        self.assertEqual(mock_kill.call_count, 4)
        self.assertEqual(contents, [str(zfs_device.lock.pid)])
        self.assertEqual(ZfsDevice.lock_refcount.get(zfs_device.lock_unique_id), 1)

    def test_locked_nonexistent_pid(self):
        """ check lock is broken when owner pid is not running """
        zfs_device = self._create_owned_lockfile()

        no_such_process_exception = OSError()
        no_such_process_exception.errno = ESRCH
        mock_kill = mock.Mock()
        mock_kill.side_effect = no_such_process_exception

        with open(zfs_device.lock.lock_file) as f:
            contents = f.readlines()

        self.assertNotEqual(contents, [str(zfs_device.lock.pid)])

        with mock.patch('os.kill', mock_kill):
            self.assertFalse(zfs_device.lock.i_am_locking())
            zfs_device.lock_pool()

        self.assertTrue(zfs_device.lock.i_am_locking())

        with open(zfs_device.lock.lock_file) as f:
            contents = f.readlines()

        self.assertEqual(mock_kill.call_count, 1)
        self.assertEqual(contents, [str(zfs_device.lock.pid)])
        self.assertEqual(ZfsDevice.lock_refcount.get(zfs_device.lock_unique_id), 1)

    def test_not_initialised(self):
        """ check lock directory is initialised """
        ZfsDevice.locks_dir_initialized = False
        zfs_device = ZfsDevice(self.zpool_name, False)

        mock_makedirs = mock.Mock()
        with mock.patch('os.makedirs', mock_makedirs):
            zfs_device.lock_pool()

            mock_makedirs.assert_called_once_with(ZfsDevice.ZPOOL_LOCK_DIR)
            mock_makedirs.reset_mock()

            zfs_device.lock_pool()

            mock_makedirs.assert_not_called()

        ZfsDevice.locks_dir_initialized = False



import mock
from django.utils import unittest
from iml_common.lib.ntp import NTPConfig


class TestNTPConfig(unittest.TestCase):
    """ Test class for NTPConfig """

    def setUp(self):
        super(TestNTPConfig, self).setUp()

        self.ntp = NTPConfig()

        # Static assignments for use in tests
        self.existing_server = 'iml.ntp.com'
        self.existing_directive = self._directive(self.existing_server)

        # Constants for retrieving from legacy IML config files
        self.existing_directive_old = 'server ' + self.existing_server
        self.manager_marker = '# Added by chroma-manager\n'
        self.manager_comment = '# Commented out by chroma-manager: '

        # each time we run _reset_and_read_conf, config file edits are removed and config reset
        self.mock_open = mock.mock_open()
        self.mock_open.return_value.readlines.return_value = self._get_lines('IML')
        mock.patch('__builtin__.open', self.mock_open, create=True).start()

        mock.patch.object(self.ntp, '_write_conf', return_value=None).start()

        self.addCleanup(mock.patch.stopall)

    @staticmethod
    def _directive(hostname):
        """ helper method to return config file server directive """
        return 'server {0} iburst'.format(hostname)

    def _get_lines(self, variant):
        lines = {
            'no-servers': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                '#server 0.centos.pool.ntp.org iburst\n',
                '#server 1.centos.pool.ntp.org iburst\n',
                '#server 2.centos.pool.ntp.org iburst\n',
                '#server 3.centos.pool.ntp.org iburst\n',
                '\n'],
            'pre-IML': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                'server 0.centos.pool.ntp.org iburst\n',
                'server 1.centos.pool.ntp.org iburst\n',
                'server 2.centos.pool.ntp.org iburst\n',
                'server 3.centos.pool.ntp.org iburst\n',
                '\n'],
            'IML': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                '{0} {1} server 0.centos.pool.ntp.org iburst\n'.format(
                    self.existing_directive, self.ntp.MARKER),
                '{0} server 1.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '{0} server 2.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '{0} server 3.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '\n'],
            'IML-append': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                '#server 0.centos.pool.ntp.org iburst\n',
                '#server 1.centos.pool.ntp.org iburst\n',
                '#server 2.centos.pool.ntp.org iburst\n',
                '#server 3.centos.pool.ntp.org iburst\n',
                '\n',
                '{0} {1} \n'.format(self.existing_directive, self.ntp.MARKER)],
            'local-ip-insert': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                'server  127.127.1.0 {0}\n'.format(self.ntp.MARKER),
                'fudge   127.127.1.0 stratum 10 {0} server 0.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '{0} server 1.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '{0} server 2.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '{0} server 3.centos.pool.ntp.org iburst\n'.format(self.ntp.MARKER),
                '\n'],
            'local-ip-append': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                '#server 0.centos.pool.ntp.org iburst\n',
                '#server 1.centos.pool.ntp.org iburst\n',
                '#server 2.centos.pool.ntp.org iburst\n',
                '#server 3.centos.pool.ntp.org iburst\n',
                '\n',
                'server  127.127.1.0 {0}\n'.format(self.ntp.MARKER),
                'fudge   127.127.1.0 stratum 10 {0} \n'.format(self.ntp.MARKER)],
            'IML-manager-old': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                self.manager_marker,
                self.existing_directive_old,
                self.manager_marker,
                '{0}server 0.centos.pool.ntp.org iburst\n'.format(self.manager_comment),
                '{0}server 1.centos.pool.ntp.org iburst\n'.format(self.manager_comment),
                '{0}server 2.centos.pool.ntp.org iburst\n'.format(self.manager_comment),
                '{0}server 3.centos.pool.ntp.org iburst\n'.format(self.manager_comment),
                '\n'],
            'local-ip-insert-manager-old': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                '#server 0.centos.pool.ntp.org iburst\n',
                '#server 1.centos.pool.ntp.org iburst\n',
                '#server 2.centos.pool.ntp.org iburst\n',
                '#server 3.centos.pool.ntp.org iburst\n',
                '#fudge \n',
                self.manager_marker,
                'server  127.127.1.0    # local clock\n',
                'fudge   127.127.1.0 stratum 10\n',
                self.manager_marker,
                '\n'],
            'local-ip-append-manager-old': [
                '# Use public servers from the pool.ntp.org project.\n',
                '# Please consider joining the pool (http://www.pool.ntp.org/join.html).\n',
                self.manager_marker,
                'server  127.127.1.0    # local clock\n',
                'fudge   127.127.1.0 stratum 10\n',
                self.manager_marker,
                '# Enable public key cryptography.',
                '\n']}
        return lines[variant]

    def test_get_server(self):
        """
        test fetching server hostname from ntp config,
        test possible sentinel and comment strings in config for backward compatibility

        because get_configured_server() doesn't use _{read|write}_conf as add() does,
        we need to mock open explicitly and make the return file object return a specific
        list of strings on readlines() call
        """
        server = self.ntp.get_configured_server(markers=None)
        self.assertEqual(server, self.existing_server)

        # now test with config with appended server directive
        self.mock_open.return_value.readlines.return_value = self._get_lines('IML-append')

        server = self.ntp.get_configured_server(markers=None)
        self.assertEqual(server, self.existing_server)

        # now test with old style config syntax for compatibility with legacy IML config files
        self.mock_open.return_value.readlines.return_value = self._get_lines('IML-manager-old')

        server = self.ntp.get_configured_server(markers=[self.manager_marker])
        self.assertEqual(server, self.existing_server)

    def test_get_server_localhost(self):
        """
        test fetching server when specified as localhost in ntp config
        test possible sentinel and comment strings in config for backward compatibility

        because get_configured_server() doesn't use _{read|write}_conf as add() does,
        we need to mock open explicitly and make the return file object return a specific
        list of strings on readlines() call
        """
        self.mock_open.return_value.readlines.return_value = self._get_lines('local-ip-insert')

        server = self.ntp.get_configured_server(markers=None)
        self.assertEqual(server, 'localhost')

        self.mock_open.return_value.readlines.return_value = self._get_lines('local-ip-insert-manager-old')

        server = self.ntp.get_configured_server(markers=[self.manager_marker])
        self.assertEqual(server, 'localhost')

        self.mock_open.return_value.readlines.return_value = self._get_lines('local-ip-append')

        server = self.ntp.get_configured_server(markers=None)
        self.assertEqual(server, 'localhost')

        self.mock_open.return_value.readlines.return_value = self._get_lines('local-ip-append-manager-old')

        server = self.ntp.get_configured_server(markers=[self.manager_marker])
        self.assertEqual(server, 'localhost')

    def test_get_server_from_empty_config(self):
        """
        test fetching server from ntp config with no active entry
        test possible sentinel and comment strings in config for backward compatibility

        because get_configured_server() doesn't use _{read|write}_conf as add() does,
        we need to mock open explicitly and make the return file object return a specific
        list of strings on readlines() call
        """
        self.mock_open.return_value.readlines.return_value = self._get_lines('pre-IML')

        server = self.ntp.get_configured_server(markers=None)
        self.assertEqual(server, None)

        server = self.ntp.get_configured_server(markers=[self.manager_marker])
        self.assertEqual(server, None)

    def test_add_remove_configured(self):
        """ test adding to un-configured content and then removing IML configurations """
        self.mock_open.return_value.readlines.return_value = self._get_lines('pre-IML')

        # add desired line to config
        error = self.ntp.add(self.existing_server)
        self.assertEqual(error, None)
        self.assertListEqual(self.ntp.lines, self._get_lines('IML'))

        # remove any IML configuration
        error = self.ntp.add(None)
        self.assertEqual(error, None)
        self.assertListEqual(self.ntp.lines, self._get_lines('pre-IML'))

    def test_add_remove_append_configured(self):
        """
        test adding and then removing IML configuration from ntp config content with commented server directives
        by removing previous configuration
        """
        self.mock_open.return_value.readlines.return_value = self._get_lines('no-servers')

        # add desired line to config
        error = self.ntp.add(self.existing_server)
        self.assertEqual(error, None)
        self.assertListEqual(self.ntp.lines, self._get_lines('IML-append'))

        # remove any IML configuration
        error = self.ntp.add(None)
        self.assertEqual(error, None)
        self.assertListEqual(self.ntp.lines, self._get_lines('no-servers'))

    def test_add_localhost(self):
        """
        test adding localhost to ntp config content
        if server directives exist, replace the first one with local clock fudge,
        if no server directives exist, append local clock fudge to end of file
        """
        self.mock_open.return_value.readlines.return_value = self._get_lines('pre-IML')

        self.ntp.add('localhost')
        self.assertListEqual(self.ntp.lines, self._get_lines('local-ip-insert'))

    def test_add_localhost_empty_config(self):
        """
        test adding localhost to ntp config content with no active server directives
        IP address and local clock 'fudge' directives should be appended when applying localhost
        """
        self.mock_open.return_value.readlines.return_value = self._get_lines('no-servers')

        self.ntp.add('localhost')
        self.assertListEqual(self.ntp.lines, self._get_lines('local-ip-append'))

    def test_add_and_reset_IML_edits(self):
        """
        test resetting previous IML edits in config file contents then adding again
        """
        for lines_out, lines_in, server in [('IML', 'pre-IML', self.existing_server),
                                            ('IML-append', 'no-servers', self.existing_server),
                                            ('local-ip-insert', 'pre-IML', 'localhost'),
                                            ('local-ip-append', 'no-servers', 'localhost')]:
            self.mock_open.return_value.readlines.return_value = self._get_lines(lines_in)

            self.ntp.add(server)

            self.assertListEqual(self.ntp.lines, self._get_lines(lines_out))

            self.mock_open.return_value.readlines.return_value = self._get_lines(lines_out)

            self.ntp.add(None)

            self.assertListEqual(self.ntp.lines, self._get_lines(lines_in))



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import time
import platform
import abc
from collections import defaultdict
from collections import namedtuple

from . import util
from . import shell


class ServiceControl(object):
    class_override = None
    __metaclass__ = abc.ABCMeta

    # dict to map service names to registered listener callbacks functions
    callbacks = defaultdict(list)

    ActionCode = namedtuple('ActionCode', ['before', 'after', 'error'])

    ServiceState = util.enum('SERVICESTARTING', 'SERVICESTARTED', 'SERVICESTARTERROR',
                             'SERVICESTOPPING', 'SERVICESTOPPED', 'SERVICESTOPERROR')

    def __init__(self, service_name):
        self.service_name = service_name

    @classmethod
    def create(cls, service_name):
        try:
            required_class = next(
                class_ for class_ in util.all_subclasses(cls) if class_._applicable())

            return required_class(service_name)
        except StopIteration:
            raise RuntimeError('Current platform not supported by any ServiceControl class')

    def _retried_action(self,
                        action,
                        action_codes,
                        validate,
                        error_message,
                        retry_count,
                        retry_time,
                        validate_time):
        """Try action fixed number of times and notify registered listeners with action
        before, after and error codes
        """
        error = None
        self.notify(self.service_name, action_codes.before)

        while retry_count > -1:
            error = action()

            if error is None:
                time.sleep(validate_time)

                # action completed successfully, validate the change
                if validate():
                    self.notify(self.service_name, action_codes.after)
                    return None

                # validation failed so populate error message
                error = error_message % self.service_name
            retry_count -= 1
            time.sleep(retry_time)

        # notify error if retries exceeded without success
        self.notify(self.service_name, action_codes.error)

        return error

    def start(self, retry_count=5, retry_time=1, validate_time=5):
        return self._retried_action(lambda: self._start(),
                                    self.ActionCode(self.ServiceState.SERVICESTARTING,
                                                    self.ServiceState.SERVICESTARTED,
                                                    self.ServiceState.SERVICESTARTERROR),
                                    lambda: self.running,
                                    'Service %s is not running after being started',
                                    retry_count,
                                    retry_time,
                                    validate_time)

    def stop(self, retry_count=5, retry_time=1, validate_time=5):
        return self._retried_action(lambda: self._stop(),
                                    self.ActionCode(self.ServiceState.SERVICESTOPPING,
                                                    self.ServiceState.SERVICESTOPPED,
                                                    self.ServiceState.SERVICESTOPERROR),
                                    lambda: not self.running,
                                    'Service %s is still running after being stopped',
                                    retry_count,
                                    retry_time,
                                    validate_time)

    def restart(self, retry_count=5, retry_time=1, validate_time=1):
        error = self.stop(retry_count, retry_time, validate_time)
        if error is None:
            error = self.start(retry_count, retry_time, validate_time)

        return error

    @classmethod
    def register_listener(cls, service, callback):
        """register a listening callback function to receive event notifications for a
        particular service's control events, applied through ServiceControl class instances
        """
        if not hasattr(callback, '__call__'):
            raise RuntimeError('callback parameter provided is not callable')
        # don't want duplicates in list of registered callbacks for a particular service
        if callback not in cls.callbacks[service]:
            cls.callbacks[service].append(callback)
        else:
            raise RuntimeError('callback already registered for service %s (%s)' % (service,
                                                                                    str(callback)))

    @classmethod
    def unregister_listener(cls, service, callback):
        """unregister a listening callback function to stop receive event notifications for a
        particular service's control events, applied through ServiceControl class instances
        """
        if not hasattr(callback, '__call__'):
            raise RuntimeError('callback parameter provided is not callable')
        # don't clear other callbacks registered for this service
        try:
            cls.callbacks[service].remove(callback)
        except ValueError:
            raise RuntimeError('callback function is not registered for service %s and cannot be '
                               'removed (%s)' % (service, str(callback)))

    @classmethod
    def notify(cls, service, action_code):
        """call registered callbacks for a particular service/action pair"""
        for callback in cls.callbacks[service]:
            callback(service, action_code)

    @abc.abstractproperty
    def running(self):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def daemon_reload(cls):
        """
        Only really applicable to systemctl, this cause the os system to reload the configuration of
        the controls, causes new or modified files to be re-read.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def enabled(self):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _applicable(cls):
        raise NotImplementedError

    @abc.abstractmethod
    def enable(self):
        raise NotImplementedError

    @abc.abstractmethod
    def reload(self):
        raise NotImplementedError

    @abc.abstractmethod
    def disable(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _start(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _stop(self):
        raise NotImplementedError


class ServiceControlEL6(ServiceControl):

    platform_use = '6'

    @property
    def running(self):
        # Returns True if the service is running. "service servicename status"
        return shell.Shell.run(['/sbin/service', self.service_name, 'status']).rc == 0

    @property
    def enabled(self):
        # Returns True if the service is enabled. "chkconfig servicename"
        return shell.Shell.run(['/sbin/chkconfig', self.service_name]).rc == 0

    @classmethod
    def _applicable(cls):
        return util.platform_info.system == 'Linux' and \
               cls.platform_use == util.platform_info.distro_version_full.split('.')[0]

    @classmethod
    def daemon_reload(cls):
        pass

    def enable(self):
        return shell.Shell.run_canned_error_message(['/sbin/chkconfig', '--add', self.service_name]) or \
               shell.Shell.run_canned_error_message(['/sbin/chkconfig', self.service_name, 'on'])

    def reload(self):
        return shell.Shell.run_canned_error_message(['/sbin/service', self.service_name, 'reload'])

    def disable(self):
        return shell.Shell.run_canned_error_message(['/sbin/chkconfig', self.service_name, 'off'])

    def _start(self):
        return shell.Shell.run_canned_error_message(['/sbin/service', self.service_name, 'start'])

    def _stop(self):
        return shell.Shell.run_canned_error_message(['/sbin/service', self.service_name, 'stop'])


class ServiceControlEL7(ServiceControl):

    platform_use = '7'

    @property
    def running(self):
        # Returns True if the service is running.
        return shell.Shell.run(['systemctl', 'is-active', self.service_name]).rc == 0

    @property
    def enabled(self):
        # Returns True if the service is enabled.
        return shell.Shell.run(['systemctl', 'is-enabled', self.service_name]).rc == 0

    @classmethod
    def _applicable(cls):
        return util.platform_info.system == 'Linux' and \
               cls.platform_use == util.platform_info.distro_version_full.split('.')[0]

    @classmethod
    def daemon_reload(cls):
        shell.Shell.run(['systemctl', 'daemon-reload'])

    def enable(self):
        return shell.Shell.run_canned_error_message(['systemctl', 'enable', self.service_name])

    def reload(self):
        return shell.Shell.run_canned_error_message(['systemctl', 'reload', self.service_name])

    def disable(self):
        return shell.Shell.run_canned_error_message(['systemctl', 'disable', self.service_name])

    def _start(self):
        return shell.Shell.run_canned_error_message(['systemctl', 'start', self.service_name])

    def _stop(self):
        return shell.Shell.run_canned_error_message(['systemctl', 'stop', self.service_name])


class ServiceControlOSX(ServiceControlEL6):
    """ Just a stub class so that running on OSX things can be made to work.

    We make OSX behave like EL6 because that historically is what happened. So
    this class provides for backwards compatibility.
    """
    @classmethod
    def _applicable(cls):
        return platform.system() == 'Darwin'



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


from blockdevice_linux import BlockDeviceLinux
from ..lib import shell


class BlockDeviceLvmVolume(BlockDeviceLinux):
    _supported_device_types = ['lvm_volume']

    @property
    def uuid(self):
        out = shell.Shell.try_run(["lvs", "--noheadings", "-o", "lv_uuid", self._device_path])

        return out.strip()



import mock

from iml_common.filesystems.filesystem_zfs import FileSystemZfs
from iml_common.blockdevices.blockdevice_zfs import BlockDeviceZfs, ZfsDevice
from iml_common.test.command_capture_testcase import CommandCaptureTestCase, CommandCaptureCommand


class TestFileSystemZFS(CommandCaptureTestCase):
    def setUp(self):
        super(TestFileSystemZFS, self).setUp()

        self.uuid_prop_mock = mock.PropertyMock(return_value='123456789123')
        self.uuid_patcher = mock.patch.object(BlockDeviceZfs, 'uuid', self.uuid_prop_mock)
        self.uuid_patcher.start()

        self.type_prop_mock = mock.PropertyMock(return_value='zfs')
        self.type_patcher = mock.patch.object(BlockDeviceZfs, 'filesystem_type', self.type_prop_mock)
        self.type_patcher.start()

        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.lock_pool').start()
        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.unlock_pool').start()

        self.patch_init_modules = mock.patch.object(FileSystemZfs, '_initialize_modules')
        self.patch_init_modules.start()

        self.filesystem = FileSystemZfs('zfs', 'zpool1')

        self.addCleanup(mock.patch.stopall)

    def test_mkfs_options(self):
        self.assertEqual([], self.filesystem.mkfs_options(None))

    def test_devices_match(self):
        self.uuid_patcher.stop()
        self.type_patcher.stop()
        mock_zfs_device = mock.MagicMock(autospec=ZfsDevice)
        mock_zfs_device.available.return_value = True
        mock_ZfsDevice = mock.Mock(return_value=mock_zfs_device)
        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice', mock_ZfsDevice).start()

        self.add_commands(CommandCaptureCommand(('zfs', 'get', '-H', '-o', 'value', 'guid', 'zpool1'),
                                                stdout='123456789123\n'))

        self.assertTrue(self.filesystem.devices_match('zpool1', 'zpool1', '123456789123'))

        self.assertRanAllCommandsInOrder()

    def test_mkfs_no_options(self):
        """
        Check --mkfsoptions are added to list of parameters provided to mkfs.lustre command when no parameters
        have been provided.
        """
        self.add_commands(CommandCaptureCommand(('zpool', 'set', 'failmode=panic', 'zpool1')),
                          CommandCaptureCommand(('mkfs.lustre', '--mkfsoptions="mountpoint=none"', 'zpool1/MGS')))

        self.assertEqual(self.filesystem.mkfs('MGS', self.filesystem.mkfs_options(None)),
                         {'uuid': '123456789123', 'filesystem_type': 'zfs', 'inode_size': None, 'inode_count': None})

        self.assertRanAllCommandsInOrder()

        self.uuid_prop_mock.assert_called_once()
        self.type_prop_mock.assert_called_once()

    def test_mkfs_no_mkfsoptions(self):
        """
        Check --mkfsoptions are added to list of parameters provided to mkfs.lustre command when some parameters have
        been provided.
        """
        self.add_commands(CommandCaptureCommand(('zpool', 'set', 'failmode=panic', 'zpool1')),
                          CommandCaptureCommand(('mkfs.lustre',
                                                 '--mgs',
                                                 '--backfstype=zfs',
                                                 '--mkfsoptions="mountpoint=none"',
                                                 'zpool1/MGS')))

        mkfs_options = ['--mgs', '--backfstype=zfs']

        self.assertEqual(self.filesystem.mkfs('MGS', mkfs_options),
                         {'uuid': '123456789123', 'filesystem_type': 'zfs', 'inode_size': None, 'inode_count': None})

        self.assertRanAllCommandsInOrder()

        self.uuid_prop_mock.assert_called_once()
        self.type_prop_mock.assert_called_once()

    def test_mkfs_no_mountpoint_option(self):
        """
        Check mountpoint is added to --mkfsoptions when parameter has been provided without mountpoint property
        assignment.
        """
        self.add_commands(CommandCaptureCommand(('zpool', 'set', 'failmode=panic', 'zpool1')),
                          CommandCaptureCommand(('mkfs.lustre',
                                                 '--mgs',
                                                 '--backfstype=zfs',
                                                 '--mkfsoptions="utf8only=on -o mountpoint=none"',
                                                 'zpool1/MGS')))

        mkfs_options = ['--mgs', '--backfstype=zfs', '--mkfsoptions="utf8only=on"']

        self.assertEqual(self.filesystem.mkfs('MGS', mkfs_options),
                         {'uuid': '123456789123', 'filesystem_type': 'zfs', 'inode_size': None, 'inode_count': None})

        self.assertRanAllCommandsInOrder()

        self.uuid_prop_mock.assert_called_once()
        self.type_prop_mock.assert_called_once()

    def test_mkfs_multiple_mkfsoptions_no_mountpoint(self):
        """
        Check mountpoint is added to --mkfsoptions when parameter has been provided without mountpoint property
        assignment but with multiple other options.
        """
        self.add_commands(CommandCaptureCommand(('zpool', 'set', 'failmode=panic', 'zpool1')),
                          CommandCaptureCommand(('mkfs.lustre',
                                                 '--mgs',
                                                 '--backfstype=zfs',
                                                 '--mkfsoptions="listsnapshots=on -o utf8only=on -o mountpoint=none"',
                                                 'zpool1/MGS')))

        mkfs_options = ['--mgs', '--backfstype=zfs', '--mkfsoptions="listsnapshots=on -o utf8only=on"']

        self.assertEqual(self.filesystem.mkfs('MGS', mkfs_options),
                         {'uuid': '123456789123', 'filesystem_type': 'zfs', 'inode_size': None, 'inode_count': None})

        self.assertRanAllCommandsInOrder()

        self.uuid_prop_mock.assert_called_once()
        self.type_prop_mock.assert_called_once()

    def test_mkfs_multiple_mkfsoptions_with_mountpoint(self):
        """
        Check mountpoint setting is overwritten in --mkfsoptions when parameter has been provided with
        mountpoint property assignment and multiple other options.
        """
        self.add_commands(CommandCaptureCommand(('zpool', 'set', 'failmode=panic', 'zpool1')),
                          CommandCaptureCommand(('mkfs.lustre',
                                                 '--mgs',
                                                 '--backfstype=zfs',
                                                 '--mkfsoptions="listsnapshots=on -o utf8only=on -o mountpoint=none"',
                                                 'zpool1/MGS')))

        mkfs_options = ['--mgs',
                        '--backfstype=zfs',
                        '--mkfsoptions="listsnapshots=on -o mountpoint=zpool1/MGS -o utf8only=on"']

        self.assertEqual(self.filesystem.mkfs('MGS', mkfs_options),
                         {'uuid': '123456789123', 'filesystem_type': 'zfs', 'inode_size': None, 'inode_count': None})

        self.assertRanAllCommandsInOrder()

        self.uuid_prop_mock.assert_called_once()
        self.type_prop_mock.assert_called_once()



import glob
import re
import mock

from os import path
from iml_common.blockdevices.blockdevice_zfs import BlockDeviceZfs, ZfsDevice
from tests.blockdevices.blockdevice_base_tests import BaseTestBD
from tests.data import example_data
from iml_common.test.command_capture_testcase import CommandCaptureCommand


class TestBlockDeviceZfs(BaseTestBD.BaseTestBlockDevice):
    pool_name = 'zfs_pool_devdiskbyidscsi0QEMU_QEMU_HARDDISK_WDWMAP3333333'
    dataset_path = '/'.join([pool_name, 'ost_index0'])

    rpm_qi_zfs_stdout = """Name        : zfs
Version     : 0.6.5.7
Release     : 1.el7
Architecture: x86_64
Install Date: Wed 05 Oct 2016 07:59:50 PDT
Group       : System Environment/Kernel
Size        : 819698
License     : CDDL
Signature   : (none)
Source RPM  : zfs-0.6.5.7-1.el7.src.rpm
Build Date  : Sat 20 Aug 2016 09:34:50 PDT
Build Host  : onyx-1-sdg1-el7-x8664.onyx.hpdd.intel.com
Relocations : (not relocatable)
URL         : http://zfsonlinux.org/
Summary     : Commands to control the kernel modules and libraries
Description :
This package contains the ZFS command line utilities.
"""

    rpm_qi_spl_stdout = """[root@lotus-33vm18 ~]# rpm -qi spl
Name        : spl
Version     : 1.2.3.4
Release     : 1.el7
Architecture: x86_64
Install Date: Wed 05 Oct 2016 08:35:27 PDT
Group       : System Environment/Kernel
Size        : 49120
License     : GPLv2+
Signature   : (none)
Source RPM  : spl-0.6.5.7-1.el7.src.rpm
Build Date  : Sat 20 Aug 2016 09:31:43 PDT
Build Host  : onyx-1-sdg1-el7-x8664.onyx.hpdd.intel.com
Relocations : (not relocatable)
URL         : http://zfsonlinux.org/
Summary     : Commands to control the kernel modules
Description :
This package contains the commands to verify the SPL
kernel modules are functioning properly.
"""

    zpool_properties = '-\tsize\t68G\t-\n' \
                       '-\thealth\tONLINE\t-\n' \
                       '-\tcachefile\t-\t-\n' \
                       '-\treadonly\toff\t-\n'
    zpool_properties_readonly = '-\tsize\t68G\t-\n' \
                                '-\thealth\tONLINE\t-\n' \
                                '-\tcachefile\t-\t-\n' \
                                '-\treadonly\ton\t-\n'

    def setUp(self):
        super(TestBlockDeviceZfs, self).setUp()

        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.lock_pool').start()
        mock.patch('iml_common.blockdevices.blockdevice_zfs.ZfsDevice.unlock_pool').start()
        mock.patch('os.mkdir').start()
        self.patch_init_modules = mock.patch.object(BlockDeviceZfs, '_initialize_modules')
        self.patch_init_modules.start()

        self.blockdevice = BlockDeviceZfs('zfs', self.pool_name)

        self.reset_command_capture()
        self.reset_command_capture_logs()
        self.addCleanup(mock.patch.stopall)

    def test_initialize_modules(self):
        self.patch_init_modules.stop()

        self.add_commands(CommandCaptureCommand(('modprobe', 'osd_zfs')),
                          CommandCaptureCommand(('modprobe', 'zfs')))

        self.blockdevice._initialize_modules()
        self.assertTrue(self.blockdevice._modules_initialized)

        self.assertRanAllCommandsInOrder()

    def test_filesystem_type_unoccupied(self):
        # this should be called with device_path referencing a zpool
        self.add_command(('zfs', 'list', '-H', '-o', 'name', '-r', self.blockdevice._device_path),
                         stdout=self.blockdevice._device_path)

        self.assertEqual(None, self.blockdevice.filesystem_type)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_type_occupied(self):
        # this should be called with device_path referencing a zpool
        self.add_command(('zfs', 'list', '-H', '-o', 'name', '-r', self.blockdevice._device_path),
                         stdout='%(pool)s\n%(pool)s/dataset_1' % {'pool': self.blockdevice._device_path})

        self.assertEqual('zfs', self.blockdevice.filesystem_type)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_info_unoccupied(self):
        # this should be called with device_path referencing a zpool
        self.add_command(('zfs', 'list', '-H', '-o', 'name', '-r', self.blockdevice._device_path),
                         stdout=self.blockdevice._device_path)

        self.assertEqual(None, self.blockdevice.filesystem_info)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_info_occupied(self):
        # this should be called with device_path referencing a zpool
        self.add_command(('zfs', 'list', '-H', '-o', 'name', '-r', self.blockdevice._device_path),
                         stdout='%(pool)s\n%(pool)s/dataset_1' % {'pool': self.blockdevice._device_path})

        self.assertEqual("Dataset 'dataset_1' found on zpool '%s'" %
                         self.blockdevice._device_path, self.blockdevice.filesystem_info)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_info_occupied_multiple(self):
        # this should be called with device_path referencing a zpool
        self.add_command(('zfs', 'list', '-H', '-o', 'name', '-r', self.blockdevice._device_path),
                         stdout='%(pool)s\n%(pool)s/dataset_1\n%(pool)s/dataset_2' % {'pool':
                                                                                      self.blockdevice._device_path})

        self.assertEqual("Datasets 'dataset_1,dataset_2' found on zpool '%s'" %
                         self.blockdevice._device_path, self.blockdevice.filesystem_info)
        self.assertRanAllCommandsInOrder()

    def test_uuid(self):
        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'),
                                                stdout=self.blockdevice._device_path),
                          CommandCaptureCommand(('zfs', 'get', '-H', '-o', 'value', 'guid', self.blockdevice._device_path),
                                                stdout='169883839435093209\n'))

        self.assertEqual('169883839435093209', self.blockdevice.uuid)
        self.assertRanAllCommandsInOrder()

    def test_preferred_fstype(self):
        self.assertEqual('zfs', self.blockdevice.preferred_fstype)

    def test_device_type(self):
        self.assertEqual('zfs', self.blockdevice.device_type)

    def test_device_path(self):
        self.assertEqual(self.pool_name, self.blockdevice._device_path)

    def test_mgs_targets(self):
        self.assertEqual({}, self.blockdevice.mgs_targets(None))

    def test_import_success(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)
        self.add_commands(CommandCaptureCommand(('zpool', 'import', self.pool_name)))

        self.assertIsNone(self.blockdevice.import_())
        self.assertRanAllCommandsInOrder()

    def test_import_existing_readonly(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        props = re.sub(r'readonly\s+off', 'readonly    on', example_data.zpool_example_properties)

        self.add_commands(CommandCaptureCommand(('zpool', 'import', self.pool_name),
                                                rc=1,
                                                stderr="cannot import '%s': a pool with that name already exists\n"
                                                       "use the form 'zpool import <pool | id> <newpool>' to give it "
                                                       "a new name\n",
                                                executions_remaining=1),
                          CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'),
                                                stdout='%s\n' % self.pool_name),
                          CommandCaptureCommand(('zpool', 'get', '-Hp', 'all', self.pool_name),
                                                stdout=props,
                                                executions_remaining=1),
                          CommandCaptureCommand(('zpool', 'export', self.pool_name)),
                          CommandCaptureCommand(('zpool', 'import', self.pool_name),
                                                executions_remaining=1),
                          CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'),
                                                stdout='%s\n' % self.pool_name),
                          CommandCaptureCommand(('zpool', 'get', '-Hp', 'all', self.pool_name),
                                                stdout=example_data.zpool_example_properties))

        self.assertIsNone(self.blockdevice.import_())
        self.assertRanAllCommandsInOrder()

    def test_export_success(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        self.add_command(('zpool', 'export', self.pool_name))

        self.assertIsNone(self.blockdevice.export())
        self.assertRanAllCommandsInOrder()

    def test_export_missing(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        self.add_command(('zpool', 'export', self.pool_name))

        self.assertIsNone(self.blockdevice.export())
        self.assertRanAllCommandsInOrder()

    def test_property_values(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'),
                                                stdout='%s\n' % self.pool_name),
                          CommandCaptureCommand(('zfs', 'get', '-Hp', '-o', 'property,value', 'all', self.blockdevice._device_path),
                                                stdout=example_data.zfs_example_properties))

        zfs_properties = self.blockdevice.zfs_properties(False)

        self.assertEqual(zfs_properties['lustre:fsname'], 'efs')
        self.assertEqual(zfs_properties['lustre:svname'], 'efs-MDT0000')
        self.assertEqual(zfs_properties['lustre:flags'], '37')
        self.assertRanAllCommandsInOrder()

    def test_targets(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        self.add_commands(CommandCaptureCommand(('zpool', 'list', '-H', '-o', 'name'),
                                                stdout='%s\n' % self.pool_name),
                          CommandCaptureCommand(('zfs', 'get', '-Hp', '-o', 'property,value', 'all', self.blockdevice._device_path),
                                                stdout=example_data.zfs_example_properties))

        target_info = self.blockdevice.targets(None, None, None)

        self.assertIn('efs-MDT0000', target_info.names)
        self.assertEqual(target_info.params['fsname'], ['efs'])
        self.assertEqual(target_info.params['svname'], ['efs-MDT0000'])
        self.assertEqual(target_info.params['flags'], ['37'])
        self.assertRanAllCommandsInOrder()

    def test_purge_filesystem_information(self):
        self.blockdevice = BlockDeviceZfs('zfs', self.dataset_path)

        self.add_commands(CommandCaptureCommand(('lctl', 'erase_lcfg', 'testfs')))

        with mock.patch.object(glob, 'glob', return_value=['testfs-a', 'testfs-b']):
            result = self.blockdevice.purge_filesystem_configuration('testfs', None)

        self.assertEqual(result, None)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver(self):
        self.add_commands(CommandCaptureCommand(('genhostid',)),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-mount')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-mount')),
                          CommandCaptureCommand(('rpm', '-qi', 'spl'), stdout=self.rpm_qi_spl_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'spl/1.2.3.4')),
                          CommandCaptureCommand(('modprobe', 'spl')),
                          CommandCaptureCommand(('rpm', '-qi', 'zfs'), stdout=self.rpm_qi_zfs_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'zfs/0.6.5.7')),
                          CommandCaptureCommand(('modprobe', 'zfs')))

        with mock.patch.object(path, 'isfile', return_value=False):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertEqual(result, None)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_file_exists(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'status', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-mount')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-mount')),
                          CommandCaptureCommand(('rpm', '-qi', 'spl'), stdout=self.rpm_qi_spl_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'spl/1.2.3.4')),
                          CommandCaptureCommand(('modprobe', 'spl')),
                          CommandCaptureCommand(('rpm', '-qi', 'zfs'), stdout=self.rpm_qi_zfs_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'zfs/0.6.5.7')),
                          CommandCaptureCommand(('modprobe', 'zfs')))

        with mock.patch.object(path, 'isfile', return_value=True):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertEqual(result, None)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_fail_genhostid(self):
        self.add_commands(CommandCaptureCommand(('genhostid',), rc=1, stderr='sample genhostid error text'))

        with mock.patch.object(path, 'isfile', return_value=False):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertIn('sample genhostid error text', result)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_fail_dkms_spl(self):
        self.add_commands(CommandCaptureCommand(('genhostid',)),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-mount')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-mount')),
                          CommandCaptureCommand(('rpm', '-qi', 'spl'), stdout=self.rpm_qi_spl_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'spl/1.2.3.4'), rc=1, stderr='sample dkms error text'))

        with mock.patch.object(path, 'isfile', return_value=False):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertIn('sample dkms error text', result)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_fail_dkms_zfs(self):
        self.add_commands(CommandCaptureCommand(('genhostid',)),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-mount')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-mount')),
                          CommandCaptureCommand(('rpm', '-qi', 'spl'), stdout=self.rpm_qi_spl_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'spl/1.2.3.4')),
                          CommandCaptureCommand(('modprobe', 'spl')),
                          CommandCaptureCommand(('rpm', '-qi', 'zfs'), stdout=self.rpm_qi_zfs_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'zfs/0.6.5.7'), rc=1, stderr='sample dkms error text'))

        with mock.patch.object(path, 'isfile', return_value=False):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertIn('sample dkms error text', result)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_fail_modprobe_zfs(self):
        self.add_commands(CommandCaptureCommand(('genhostid',)),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs.target')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-scan')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-import-cache')),
                          CommandCaptureCommand(('systemctl', 'status', 'zfs-mount')),
                          CommandCaptureCommand(('systemctl', 'disable', 'zfs-mount')),
                          CommandCaptureCommand(('rpm', '-qi', 'spl'), stdout=self.rpm_qi_spl_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'spl/1.2.3.4')),
                          CommandCaptureCommand(('modprobe', 'spl')),
                          CommandCaptureCommand(('rpm', '-qi', 'zfs'), stdout=self.rpm_qi_zfs_stdout),
                          CommandCaptureCommand(('dkms', 'install', 'zfs/0.6.5.7')),
                          CommandCaptureCommand(('modprobe', 'zfs'), rc=1, stderr='sample modprobe error text'))

        with mock.patch.object(path, 'isfile', return_value=False):
            result = BlockDeviceZfs.initialise_driver(True)

        self.assertIn('sample modprobe error text', result)
        self.assertRanAllCommandsInOrder()

    def test_initialise_driver_monitor_mode(self):
        result = BlockDeviceZfs.initialise_driver(False)

        self.assertEqual(result, None)
        self.assertRanAllCommandsInOrder()

    def _base_terminate_driver_test(self, getpid_retval, listdir_retval, lockfilepid_retval):
        mock_remove = mock.Mock()
        mock.patch('os.remove', mock_remove).start()
        mock_getpid = mock.Mock(return_value=getpid_retval)
        mock.patch('os.getpid', mock_getpid).start()
        mock_listdir = mock.Mock(return_value=listdir_retval)
        mock.patch('os.listdir', mock_listdir).start()
        mock_lockfilepid = mock.Mock(return_value=lockfilepid_retval)
        mock.patch('iml_common.blockdevices.blockdevice_zfs.get_lockfile_pid', mock_lockfilepid).start()

        result = BlockDeviceZfs.terminate_driver()

        self.assertEqual(result, None)

        return mock_remove, mock_getpid, mock_listdir, mock_lockfilepid

    def test_terminate_driver_active_locks_this_process(self):
        mock_remove, mock_getpid, mock_listdir, mock_lockfilepid = self._base_terminate_driver_test(1234,
                                                                                                    ['pool1'],
                                                                                                    1234)
        mock_remove.assert_called_once_with('%s/pool1' % ZfsDevice.ZPOOL_LOCK_DIR)
        mock_getpid.assert_called_once_with()
        mock_listdir.assert_called_once_with(ZfsDevice.ZPOOL_LOCK_DIR)
        mock_lockfilepid.assert_called_once_with('%s/pool1' % ZfsDevice.ZPOOL_LOCK_DIR)

    def test_terminate_driver_active_locks_other_process(self):
        mock_remove, mock_getpid, mock_listdir, mock_lockfilepid = self._base_terminate_driver_test(1233,
                                                                                                    ['pool1'],
                                                                                                    1234)
        assert mock_remove.call_args_list == []
        mock_getpid.assert_called_once_with()
        mock_listdir.assert_called_once_with(ZfsDevice.ZPOOL_LOCK_DIR)
        mock_lockfilepid.assert_called_once_with('%s/pool1' % ZfsDevice.ZPOOL_LOCK_DIR)

    def test_terminate_driver_file_with_no_pid(self):
        mock_remove = mock.Mock()
        mock.patch('os.remove', mock_remove).start()
        mock_getpid = mock.Mock(return_value=1234)
        mock.patch('os.getpid', mock_getpid).start()
        mock_listdir = mock.Mock(return_value=['pool1'])
        mock.patch('os.listdir', mock_listdir).start()
        mock_lockfilepid = mock.Mock(side_effect=AssertionError)
        mock.patch('iml_common.blockdevices.blockdevice_zfs.get_lockfile_pid', mock_lockfilepid).start()

        result = BlockDeviceZfs.terminate_driver()

        self.assertEqual(result, None)

        assert mock_remove.call_args_list == []
        assert mock_getpid.call_args_list == []
        mock_listdir.assert_called_once_with(ZfsDevice.ZPOOL_LOCK_DIR)
        mock_lockfilepid.assert_called_once_with('%s/pool1' % ZfsDevice.ZPOOL_LOCK_DIR)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


from ..lib import shell
from ..blockdevices.blockdevice_zfs import BlockDeviceZfs
from filesystem import FileSystem


class FileSystemZfs(FileSystem, BlockDeviceZfs):
    _supported_filesystems = ['zfs']

    def __init__(self, fstype, device_path):
        super(FileSystemZfs, self).__init__(fstype, device_path)

        self._modules_initialized = False
        self._zfs_properties = None

    @property
    def label(self):
        return BlockDeviceZfs('zfs', self._device_path).zfs_properties(False)['lustre:svname']

    @property
    def inode_size(self):
        # TODO: this is knowledge of the subclass specific type in the base class which shouldn't be there
        return None

    @property
    def inode_count(self):
        # TODO: this is knowledge of the subclass specific type in the base class which shouldn't be there
        return None

    def mount_path(self, target_name):
        """
        Before FormatTargetJob, _device_path will reference the zpool, but afterwards
        _device_path will reference the zpool/dataset (after being updated during
        UpdateManagedTargetMount step)

        :param target_name: lustre target name
        :return: lustre target path <zpool>/<dataset>
        """
        return "%s/%s" % (self._device_path, target_name)

    def mkfs(self, target_name, options):
        """
        Retrieve new target dataset path and pass in call to mkfs.lustre.
        Set failmode to panic for underlying zpool and ensure correct property values set when creating dataset by
        supplying mkfsoptions which are passed to the underlying backfs tools (zfs in this case).

        :param target_name: lustre target name
        :param options: list of options to supply to mkfs command
        :return: dict of new target info
        """
        self._initialize_modules()

        # set 'failmode=panic' property on underlying device (zpool)
        BlockDeviceZfs('zfs', self._device_path).failmode = 'panic'

        new_path = self.mount_path(target_name)

        # set 'mountpoint=none' for created ZfsDatasets
        try:
            options_idx = next(options.index(option) for option in options if '--mkfsoptions=' in option)
        except StopIteration:
            # no mkfsoptions option exists, add one
            options.append('--mkfsoptions="mountpoint=none"')
        else:
            # retrieve list of mkfsoptions from existing parameter string
            mkfsoptions = options[options_idx].split('=', 1)[1].strip('"').split(' -o ')
            try:
                mountpoint_idx = next(mkfsoptions.index(option) for option in mkfsoptions if 'mountpoint=' in option)
            except StopIteration:
                pass
            else:
                # we want to overwrite any existing mountpoint property value, so first remove
                mkfsoptions.pop(mountpoint_idx)

            mkfsoptions.append('mountpoint=none')
            options[options_idx] = '--mkfsoptions="%s"' % ' -o '.join([opt for opt in mkfsoptions])

        shell.Shell.try_run(["mkfs.lustre"] + options + [new_path])

        return {'uuid': BlockDeviceZfs('zfs', new_path).uuid,
                'filesystem_type': self.filesystem_type,
                'inode_size': None,
                'inode_count': None}

    def mkfs_options(self, target):
        return []

    def devices_match(self, device1_path, device2_path, device2_uuid):
        """
        Verifies that the devices referenced in the parameters are the same

        :param device1_path: first device string representation
        :param device2_path: second device string representation
        :param device2_uuid: uuid of second device
        :return: return True if both device identifiers reference the same object
        """
        return device2_uuid == BlockDeviceZfs('zfs', device1_path).uuid



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


# import so that this module can be imported in pure python environments as well as on Linux.
# Doing it this way means that rpm_lib can mocked.
try:
    import rpm as rpm_lib
except:
    try:
        # Allow an override in local settings, because I want to get real output I run it on a remote linux node from my mac.
        # I have this in my local_settings.py : Chris
        # class rpm_lib(object):
        #     @classmethod
        #     def labelCompare(cls, a, b):
        #         from subprocess import Popen, PIPE
        #         result = Popen(["ssh", "lotus-33",  "python -c \"import rpm\nimport sys\nsys.stdout.write(str(rpm.labelCompare(('%s', '%s', '%s'), ('%s', '%s', '%s'))))\"" % (a[0], a[1], a[2], b[0], b[1], b[2])], stdout=PIPE).communicate()[0]
        #         return int(result
        from local_settings import rpm_lib
    except:
        class rpm_lib(object):
            @classmethod
            def labelCompare(cls, a, b):
                return cmp(a, b)


class VersionInfo(object):
    """
    A convenient way of storing package version information that can be printed and compared with ease.

    At present this class is not serializable and so cannot be converted easily into json. This highlights
    a limitation of our current use of plain json for messages in that all but the basic types have to be
    discarded during their transportation.
    """
    def __init__(self, epoch, version, release, arch):
        self.epoch = epoch
        self.version = version
        self.release = release
        self.arch = arch

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "epoch='%s', version='%s', release='%s', arch='%s'" % (self.epoch, self.version, self.release, self.arch)

    def __cmp__(self, other):
        return rpm_lib.labelCompare((self.epoch, self.version, self.release), (other.epoch, other.version, other.release))



""" Data for use in iml_common unit tests, mostly example commandline stdout """

# example commandline stdout from command of the format:
# >> zfs get -Hp -o property,value all <zfs dataset path e.g. pool1/MGS>
zfs_example_properties = """type	filesystem
garbageline followed by blank line

creation	1412156944
used	4602880
available	21002432000
referenced	4602880
compressratio	1.00x
mounted	no
quota	0
reservation	0
recordsize	131072
mountpoint	/zfs_pool_devdiskbyidscsi0QEMU_QEMU_HARDDISK_WDWMAP2222222/mgt
sharenfs	off
checksum	on
compression	off
atime	on
devices	on
exec	on
setuid	on
readonly	off
zoned	off
snapdir	hidden
aclinherit	restricted
canmount	off
xattr	sa
copies	1
version	5
utf8only	off
normalization	none
casesensitivity	sensitive
vscan	off
nbmand	off
sharesmb	off
refquota	0
refreservation	0
primarycache	all
secondarycache	all
usedbysnapshots	0
usedbydataset	4602880
usedbychildren	0
usedbyrefreservation	0
logbias	latency
dedup	off
mlslabel	none
sync	standard
refcompressratio	1.00x
written	4602880
logicalused	3874304
logicalreferenced	3874304
snapdev	hidden
acltype	off
context	none
fscontext	none
defcontext	none
rootcontext	none
relatime	off
lustre:version	1
lustre:fsname	efs
lustre:index	0
lustre:svname	efs-MDT0000
lustre:flags	37"""

# example commandline stdout from command of the format:
# >> dumpe2fs -h <ldiskfs device path e.g. /dev/sda1>
dumpe2fs_example_output = """
Filesystem volume name:   <none>
Last mounted on:          /boot
Filesystem UUID:          e77433f9-d9fb-48e7-b444-41a3c84016db
Filesystem magic number:  0xEF53
Filesystem revision #:    1 (dynamic)
Filesystem features:      has_journal ext_attr resize_inode dir_index filetype needs_recovery sparse_super
Filesystem flags:         signed_directory_hash
Default mount options:    user_xattr acl
Filesystem state:         clean
Errors behavior:          Continue
Filesystem OS type:       Linux
Inode count:              128016
Block count:              512000
Reserved block count:     25600
Free blocks:              308769
Free inodes:              127670
First block:              1
Block size:               1024
Fragment size:            1024
Reserved GDT blocks:      256
Blocks per group:         8192
Fragments per group:      8192
Inodes per group:         2032
Inode blocks per group:   254
Filesystem created:       Thu May 26 11:23:07 2016
Last mount time:          Thu Jun 30 06:21:32 2016
Last write time:          Thu Jun 30 06:21:32 2016
Mount count:              36
Maximum mount count:      -1
Last checked:             Thu May 26 11:23:07 2016
Check interval:           0 (<none>)
Lifetime writes:          1425 MB
Reserved blocks uid:      0 (user root)
Reserved blocks gid:      0 (group root)
First inode:              11
Inode size:               128
Journal inode:            8
Default directory hash:   half_md4
Directory Hash Seed:      0d04e737-aece-4ef7-9e21-1edea14ebcde
Journal backup:           inode blocks
Journal features:         journal_incompat_revoke
Journal size:             8M
Journal length:           8192
Journal sequence:         0x00000365
Journal start:            1"""

# example commandline stdout from command of the format:
# >> zpool get -Hp all <zpool name>
zpool_example_properties = """bob     size    10670309376     -
bob     capacity        0       -
bob     altroot -       default
bob     health  ONLINE  -
bob     guid    14447627934634597108    -
bob     version -       default
bob     bootfs  -       default
bob     delegation      on      default
bob     autoreplace     off     default
bob     cachefile       -       default
bob     failmode        panic   local
bob     listsnapshots   off     default
bob     autoexpand      off     default
bob     dedupditto      0       default
bob     dedupratio      1.00    -
bob     free    10670149632     -
bob     allocated       159744  -
bob     readonly        off     -
bob     ashift  0       default
bob     comment -       default
bob     expandsize      -       -
bob     freeing 0       -
bob     fragmentation   0       -
bob     leaked  0       -
bob     multihost       on      local
bob     feature@async_destroy   enabled local
bob     feature@empty_bpobj     active  local
bob     feature@lz4_compress    active  local
bob     feature@multi_vdev_crash_dump   enabled local
bob     feature@spacemap_histogram      active  local
bob     feature@enabled_txg     active  local
bob     feature@hole_birth      active  local
bob     feature@extensible_dataset      active  local
bob     feature@embedded_data   active  local
bob     feature@bookmarks       enabled local
bob     feature@filesystem_limits       enabled local
bob     feature@large_blocks    enabled local
bob     feature@large_dnode     enabled local
bob     feature@sha512  enabled local
bob     feature@skein   enabled local
bob     feature@edonr   enabled local
bob     feature@userobj_accounting      active  local"""



from iml_common.blockdevices.blockdevice_lvm_volume import BlockDeviceLvmVolume
from iml_common.test.command_capture_testcase import CommandCaptureTestCase


class TestBlockDeviceLvmVolume(CommandCaptureTestCase):
    def setUp(self):
        super(TestBlockDeviceLvmVolume, self).setUp()

        self.blockdevice = BlockDeviceLvmVolume('lvm_volume', '/dev/mappper/lvg-test-lvm-test')

    def test_uuid(self):
        self.add_command(("lvs", "--noheadings", "-o", "lv_uuid", '/dev/mappper/lvg-test-lvm-test'), stdout='  CtSfyh-ThdO-Bg3i-EiKU-6knJ-Ix4D-ru49Py\n')

        self.assertEqual('CtSfyh-ThdO-Bg3i-EiKU-6knJ-Ix4D-ru49Py', self.blockdevice.uuid)



import mock

from django.utils import unittest
from iml_common.lib.exception_sandbox import exceptionSandBox


class FencingTestCase(unittest.TestCase):

    mock_logger = mock.Mock()

    @exceptionSandBox(mock_logger, 1)
    def function1(self):
        if self.raise_exception:
            raise Exception()

        return 2

    def test_no_exception_debug(self):
        self.raise_exception = False

        self.assertEqual(self.function1(), 2)

    def test_exception_debug(self):
        self.raise_exception = True

        self.assertEqual(self.function1(), 1)
        self.mock_logger.debug.assert_called_once()



import json

from django.utils import unittest

from iml_common.lib.name_value_list import NameValueList


class TestNameValueList(unittest.TestCase):

    def setUp(self):
        self.raw_values = [{'resolve': False},
                           {'ping': True},
                           {'auth': False},
                           {'hostname_valid': False},
                           {'fqdn_resolves': False},
                           {'fqdn_matches': False},
                           {'reverse_resolve': False},
                           {'reverse_ping': False},
                           {'yum_valid_repos': False},
                           {'yum_can_update': False},
                           {'openssl': False}]

        self.values = NameValueList(self.raw_values)

    def test_creation_correction(self):
        for value, raw_value in zip(self.values, self.raw_values):
            self.assertEqual(value.name, raw_value.keys()[0])
            self.assertEqual(value.value, raw_value.values()[0])

    def test_name_value_create(self):
        self.values['execute'] = True
        self.assertEqual(self.values['execute'], True)
        pass

    def test_update(self):
        self.values['resolve'] = True
        self.assertEqual(self.values['resolve'], True)

    def test_type_change(self):
        self.values['reverse_resolve'] = 'string'
        self.assertEqual(self.values['reverse_resolve'], 'string')

    def test_index_constant(self):
        self.values['reverse_resolve'] = True
        self.assertEqual(self.values.keys().index('reverse_resolve'), 6)

    def test_collection_via_json(self):
        json_string = json.dumps(self.values.collection())
        loaded_values = NameValueList(json.loads(json_string))
        self.assertEqual(self.values, loaded_values)

    def test_collection_len(self):
        self.assertEqual(len(self.values), len(self.values.collection()))

    def test_collection_values(self):
        for entry in self.values.collection():
            self.assertEqual(entry['value'], self.values[entry['name']])



import mock

from iml_common.test.command_capture_testcase import CommandCaptureTestCase, CommandCaptureCommand
from iml_common.filesystems.filesystem_ldiskfs import FileSystemLdiskfs
from iml_common.blockdevices.blockdevice_linux import BlockDeviceLinux
from tests.data import example_data


class TestFileSystemLdiskfs(CommandCaptureTestCase):
    def setUp(self):
        super(TestFileSystemLdiskfs, self).setUp()

        self.uuid_prop_mock = mock.PropertyMock(return_value='123456789123')
        mock.patch.object(BlockDeviceLinux, 'uuid', self.uuid_prop_mock).start()

        self.type_prop_mock = mock.PropertyMock(return_value='ldiskfs')
        mock.patch.object(BlockDeviceLinux, 'filesystem_type', self.type_prop_mock).start()

        self.patch_init_modules = mock.patch.object(FileSystemLdiskfs, '_initialize_modules')
        self.patch_init_modules.start()

        self.filesystem = FileSystemLdiskfs('ldiskfs', '/dev/sda1')

        self.addCleanup(mock.patch.stopall)

    def test_initialize_modules(self):
        self.patch_init_modules.stop()

        self.add_commands(CommandCaptureCommand(('modprobe', 'osd_ldiskfs'), rc=1),
                          CommandCaptureCommand(('modprobe', 'ldiskfs')))

        self.filesystem._initialize_modules()
        self.assertTrue(self.filesystem._modules_initialized)

        self.assertRanAllCommandsInOrder()

    def _mount_fail_initial(self, fail_code):
        """ Test when initial mount fails, retry succeeds and result returned """
        self.add_commands(CommandCaptureCommand(('mount', '-t', 'lustre', '/dev/sda1', '/mnt/OST0000'), rc=fail_code,
                                                executions_remaining=1),
                          CommandCaptureCommand(('mount', '-t', 'lustre', '/dev/sda1', '/mnt/OST0000'), rc=0,
                                                executions_remaining=1))

        self.filesystem.mount('/mnt/OST0000')

        self.assertRanAllCommandsInOrder()

    def test_mount_fail_initial_5(self):
        self._mount_fail_initial(5)

    def test_mount_fail_initial_108(self):
        self._mount_fail_initial(108)

    def test_mount_different_rc_fail_initial(self):
        """ Test when initial mount fails and the rc doesn't cause a retry, exception is raised """
        self.add_commands(CommandCaptureCommand(('mount', '-t', 'lustre', '/dev/sda1', '/mnt/OST0000'), rc=1,
                                                executions_remaining=1))

        self.assertRaises(RuntimeError, self.filesystem.mount, '/mnt/OST0000')

        self.assertRanAllCommandsInOrder()

    def test_mount_fail_second(self):
        """ Test when initial mount fails and the retry fails, exception is raised """
        self.add_commands(CommandCaptureCommand(('mount', '-t', 'lustre', '/dev/sda1', '/mnt/OST0000'), rc=5,
                                                executions_remaining=1),
                          CommandCaptureCommand(('mount', '-t', 'lustre', '/dev/sda1', '/mnt/OST0000'), rc=5,
                                                executions_remaining=1))

        self.assertRaises(RuntimeError, self.filesystem.mount, '/mnt/OST0000')

        self.assertRanAllCommandsInOrder()

    def test_mkfs_options(self):
        test_target = mock.Mock()
        test_target.inode_size = 512
        test_target.bytes_per_inode = 256
        test_target.inode_count = 1024

        self.assertEqual(['-I 512', '-i 256', '-N 1024'], self.filesystem.mkfs_options(test_target))

    def test_devices_match(self):
        mock_stat = mock.patch('os.stat').start()
        self.filesystem.devices_match('/dev/sda1', '/dev/disk/by-id/link_to_sda1', '123456789123')

        mock_stat.assert_has_calls([mock.call('/dev/sda1'), mock.call('/dev/disk/by-id/link_to_sda1')])

    def test_mkfs(self):
        """ Test returning the correct parameters from mkfs call. Test also partly covers inode methods """
        self.add_commands(CommandCaptureCommand(('mkfs.lustre', '/dev/sda1')),
                          CommandCaptureCommand(('dumpe2fs', '-h', self.filesystem._device_path),
                                                stdout=example_data.dumpe2fs_example_output),
                          CommandCaptureCommand(('dumpe2fs', '-h', self.filesystem._device_path),
                                                stdout=example_data.dumpe2fs_example_output))

        self.assertEqual(self.filesystem.mkfs(None, []),
                         {'uuid': '123456789123',
                          'filesystem_type': 'ldiskfs',
                          'inode_size': 128,
                          'inode_count': 128016})

        self.assertRanAllCommandsInOrder()



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import mock

from django.utils import unittest


class ImlUnitTestCase(unittest.TestCase):
    def setUp(self):
        super(ImlUnitTestCase, self).setUp()

        self.addCleanup(mock.patch.stopall)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import sys
import traceback

'''
This allows code to be placed in safe space for production (?) where uncaught exceptions are captured and logged
but passed not further. This means we can have a resilient application for production but see all the badness in
test / development. Use carefully to sandbox functionality.

It is implemented as a context handler and a decorator so that the user can use the most appropriate implementation
for there needs.
'''


def exceptionSandBox(logger, exception_value, message='Exception raised in sandbox'):
    def real_decorator(function):
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except Exception:
                ex_type, ex, tb = sys.exc_info()
                data = traceback.format_exc(tb).splitlines()
                logger.debug('%s:\n%s' % (message + ' START',
                                          '\n'.join(data[1:] + [message + ' END'])))
            return exception_value
        return wrapper

    return real_decorator



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import blockdevice_linux
import blockdevice_zfs
import blockdevice_lvm_volume



import mock
from os import environ
from iml_common.lib import util
from iml_common.test.iml_unit_testcase import ImlUnitTestCase


def target_func():
    return


class PreservePermissionTestCase(ImlUnitTestCase):

    def test_file(self):
        for filename, st_mode, st_uid, st_gid in [('sidney', 493, 2, 10),
                                                  ('denver', 362, 4, 7),
                                                  ('shytown', 302, 5, 8)]:
            magic_stat_mock = mock.MagicMock()
            magic_stat_mock.st_mode = st_mode
            magic_stat_mock.st_uid = st_uid
            magic_stat_mock.st_gid = st_gid

            mock_os_stat = mock.patch('os.stat', return_value=magic_stat_mock).start()
            mock_os_chown = mock.patch('os.chown').start()
            mock_os_chmod = mock.patch('os.chmod').start()

            with util.PreserveFileAttributes(filename):
                self.assertEqual(mock_os_stat.call_count, 3)
                mock_os_stat.assert_called_with(filename)

            self.assertEqual(mock_os_chmod.call_count, 1)
            mock_os_chmod.assert_called_with(filename, magic_stat_mock.st_mode)
            self.assertEqual(mock_os_chown.call_count, 1)
            mock_os_chown.assert_called_with(filename, magic_stat_mock.st_uid, magic_stat_mock.st_gid)


class DisableThreadsTestCase(ImlUnitTestCase):

    def setUp(self):
        super(DisableThreadsTestCase, self).setUp()
        environ.pop('IML_DISABLE_THREADS', None)

    def tearDown(self):
        environ.pop('IML_DISABLE_THREADS', None)
        super(DisableThreadsTestCase, self).tearDown()

    def test_threads_on(self):
        thread = util.ExceptionThrowingThread(target=target_func, args=())

        self.assertTrue(thread._use_threads)

    def test_threads_on_values(self):
        values = ['0', '']
        for value in values:
            environ['IML_DISABLE_THREADS'] = value
            thread = util.ExceptionThrowingThread(target=target_func, args=())

            self.assertTrue(thread._use_threads)

    def test_threads_off(self):
        environ['IML_DISABLE_THREADS'] = '1'
        thread = util.ExceptionThrowingThread(target=target_func, args=())

        self.assertFalse(thread._use_threads)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


AGENTRPCWRAPPERVERSION = 1


def agent_error(value):
    return {'wrapper_version': AGENTRPCWRAPPERVERSION,
            'error': value}


def agent_result(value):
    return {'wrapper_version': AGENTRPCWRAPPERVERSION,
            'result': value}


def agent_ok_or_error(error):
    '''
    If error != None then return it as the error else return result_ok
    :param value: error or None
    :return:agent_error or agent_result_ok
    '''
    if error:
        return agent_error(error)
    else:
        return agent_result_ok

agent_result_ok = agent_result(True)


def agent_result_is_error(result):
    return 'error' in result


def agent_result_is_ok(result):
    return 'result' in result



import mock

from iml_common.blockdevices.blockdevice_linux import BlockDeviceLinux
from tests.blockdevices.blockdevice_base_tests import BaseTestBD
from iml_common.test.command_capture_testcase import CommandCaptureCommand


class TestBlockDeviceLinux(BaseTestBD.BaseTestBlockDevice):
    def setUp(self):
        super(TestBlockDeviceLinux, self).setUp()

        self.patch_init_modules = mock.patch.object(BlockDeviceLinux, '_initialize_modules')
        self.patch_init_modules.start()

        self.blockdevice = BlockDeviceLinux('linux', '/dev/sda1')

        self.addCleanup(mock.patch.stopall)

    def test_initialize_modules(self):
        self.patch_init_modules.stop()

        self.add_commands(CommandCaptureCommand(('modprobe', 'osd_ldiskfs'), rc=1),
                          CommandCaptureCommand(('modprobe', 'ldiskfs')))

        self.blockdevice._initialize_modules()
        self.assertTrue(self.blockdevice._modules_initialized)

        self.assertRanAllCommandsInOrder()

    def test_filesystem_type_unoccupied(self):
        self.add_command(('blkid', '-p', '-o', 'value', '-s', 'TYPE', self.blockdevice._device_path))

        self.assertEqual(None, self.blockdevice.filesystem_type)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_type_occupied(self):
        self.add_command(('blkid', '-p', '-o', 'value', '-s', 'TYPE', self.blockdevice._device_path),
                         stdout='ext3')

        self.assertEqual('ext3', self.blockdevice.filesystem_type)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_info_unoccupied(self):
        self.add_command(('blkid', '-p', '-o', 'value', '-s', 'TYPE', self.blockdevice._device_path))

        self.assertEqual(None, self.blockdevice.filesystem_info)
        self.assertRanAllCommandsInOrder()

    def test_filesystem_info_occupied(self):
        self.add_command(('blkid', '-p', '-o', 'value', '-s', 'TYPE', self.blockdevice._device_path),
                         stdout='ext3')

        self.assertEqual("Filesystem found: type 'ext3'", self.blockdevice.filesystem_info)
        self.assertRanAllCommandsInOrder()

    def test_uuid(self):
        self.add_command(('blkid', '-p', '-o', 'value', '-s', 'UUID', self.blockdevice._device_path))

        self.assertEqual(None, self.blockdevice.uuid)
        self.assertRanAllCommandsInOrder()

    def test_preferred_fstype(self):
        self.assertEqual('ldiskfs', self.blockdevice.preferred_fstype)

    def test_device_type(self):
        self.assertEqual('linux', self.blockdevice.device_type)

    def test_device_path(self):
        self.assertEqual('/dev/sda1', self.blockdevice._device_path)

    def test_mgs_targets(self):
        # TODO: implement test
        pass

    def test_property_values(self):
        # TODO: implement test
        pass

    def test_targets(self):
        # TODO: implement test
        pass



import tempfile
import mock
import abc

from iml_common.lib import util
from iml_common.lib.firewall_control import FirewallControl
from iml_common.lib.firewall_control import FirewallControlEL6
from iml_common.test.command_capture_testcase import CommandCaptureTestCase
from iml_common.test.command_capture_testcase import CommandCaptureCommand


class BaseTestFC:
    """ Dummy class to stop encapsulated abstract classes from being run by test runner """
    def __init__(self):
        pass

    class BaseTestFirewallControl(CommandCaptureTestCase):
        """ Abstract base class for testing the FirewallControl class """
        __metaclass__ = abc.ABCMeta

        # class variables
        port = '88'
        proto = 'tcp'
        desc = 'test service'
        address = '192.168.1.100'

        # create example named tuple to compare with objects in rules list
        example_port_rule = FirewallControl.FirewallRule(port, proto, desc, persist=True, address=None)
        example_address_rule = FirewallControl.FirewallRule(0, proto, desc, persist=False, address=address)

        # base expected error strings
        assert_address_with_port_msg = 'ing a specific port on a source address is not ' \
                                       'supported, port value must be 0 (ANY)'
        assert_address_persist_msg = 'ing all ports on a source address permanently is not ' \
                                     'currently supported'
        assert_port_not_persist_msg = 'ing a single port temporarily is not currently supported'

        def init_firewall(self, el_version):
            self.el_version = el_version

        def setUp(self):
            super(BaseTestFC.BaseTestFirewallControl, self).setUp()

            mock.patch.object(util, 'platform_info', util.PlatformInfo('Linux',
                                                                       'CentOS',
                                                                       0.0,
                                                                       self.el_version,
                                                                       0.0,
                                                                       0,
                                                                       '')).start()

            self.test_firewall = FirewallControl.create()

            # Guaranteed cleanup with unittest2
            self.addCleanup(mock.patch.stopall)

        def test_open_address_incorrect_port(self):
            # test opening a given port on a specific address, AssertionError should be
            # thrown and no commands issued
            try:
                self.test_firewall.add_rule(1234, self.proto, self.desc, persist=True, address=self.address)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'open' + self.assert_address_with_port_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        def test_close_address_incorrect_port(self):
            # test closing a given port on a specific address, AssertionError should be
            # thrown and no commands issued
            try:
                self.test_firewall.remove_rule(1234, self.proto, self.desc, persist=True, address=self.address)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'clos' + self.assert_address_with_port_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        def test_open_address_persistent(self):
            # test opening all ports on a specific address permanently, AssertionError
            # should be thrown and no commands issued
            try:
                self.test_firewall.add_rule(0, self.proto, self.desc, persist=True, address=self.address)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'open' + self.assert_address_persist_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        def test_close_address_persistent(self):
            # test closing all ports on a specific address permanently, AssertionError
            # should be thrown and no commands issued
            try:
                self.test_firewall.remove_rule(0, self.proto, self.desc, persist=True, address=self.address)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'clos' + self.assert_address_persist_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        def test_open_port_not_persistent(self):
            # test opening port on a specific address temporarily, AssertionError
            # should be thrown and no commands issued
            try:
                self.test_firewall.add_rule(1234, self.proto, self.desc, persist=False, address=None)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'open' + self.assert_port_not_persist_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        def test_close_port_not_persistent(self):
            # test closing port on a specific address temporarily, AssertionError
            # should be thrown and no commands issued
            try:
                self.test_firewall.remove_rule(1234, self.proto, self.desc, persist=False, address=None)
                # shouldn't get here
                self.assertTrue(False)
            except AssertionError as e:
                self.assertEqual(str(e), 'clos' + self.assert_port_not_persist_msg)

            # class instance should have record of added rule
            self.assertEqual(len(self.test_firewall.rules), 0)
            # no commands should have run
            self.assertRanAllCommandsInOrder()

        @abc.abstractmethod
        def test_open_port(self):
            raise NotImplementedError

        @abc.abstractmethod
        def test_close_port(self):
            raise NotImplementedError

        @abc.abstractmethod
        def test_open_address(self):
            raise NotImplementedError

        @abc.abstractmethod
        def test_close_address(self):
            raise NotImplementedError


class TestFirewallControlEL6(BaseTestFC.BaseTestFirewallControl):

    # example output from 'iptables -L' or 'service iptables status'
    chain_output = """Table: filter
Chain INPUT (policy ACCEPT)
num  target     prot opt source               destination
1    ACCEPT     all  --  0.0.0.0/0            0.0.0.0/0           state RELATED,ESTABLISHED
2    ACCEPT     icmp --  0.0.0.0/0            0.0.0.0/0
3    ACCEPT     all  --  0.0.0.0/0            0.0.0.0/0
4    ACCEPT     udp  --  0.0.0.0/0            0.0.0.0/0           state NEW udp dpt:123
5    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:22
6    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:80
7    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:443
8    REJECT     all  --  0.0.0.0/0            0.0.0.0/0           reject-with icmp-host-prohibited

Chain FORWARD (policy ACCEPT)
num  target     prot opt source               destination
1    REJECT     all  --  0.0.0.0/0            0.0.0.0/0           reject-with icmp-host-prohibited

Chain OUTPUT (policy ACCEPT)
num  target     prot opt source               destination

"""

    # example output from 'iptables -L' or 'service iptables status' with matching rule at different index
    different_index_output = """Table: filter
Chain INPUT (policy ACCEPT)
num  target     prot opt source               destination
1    ACCEPT     all  --  0.0.0.0/0            0.0.0.0/0           state RELATED,ESTABLISHED
2    ACCEPT     icmp --  0.0.0.0/0            0.0.0.0/0
3    ACCEPT     all  --  0.0.0.0/0            0.0.0.0/0
4    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:22
5    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:80
6    ACCEPT     tcp  --  0.0.0.0/0            0.0.0.0/0           state NEW tcp dpt:443
7    ACCEPT     udp  --  0.0.0.0/0            0.0.0.0/0           state NEW udp dpt:123
8    REJECT     all  --  0.0.0.0/0            0.0.0.0/0           reject-with icmp-host-prohibited

Chain FORWARD (policy ACCEPT)
num  target     prot opt source               destination
1    REJECT     all  --  0.0.0.0/0            0.0.0.0/0           reject-with icmp-host-prohibited

Chain OUTPUT (policy ACCEPT)
num  target     prot opt source               destination

"""

    # example output from 'iptables -L' or 'service iptables status' if firewall not configured
    not_configured_output = """Table: filter
Chain INPUT (policy ACCEPT)
num  target     prot opt source               destination

Chain FORWARD (policy ACCEPT)
num  target     prot opt source               destination

Chain OUTPUT (policy ACCEPT)
num  target     prot opt source               destination

"""

    not_running_msg = 'iptables: Firewall is not running.'

    def __init__(self, *args, **kwargs):
        super(TestFirewallControlEL6, self).__init__(*args, **kwargs)
        self.init_firewall('6.6')

    def test_open_port(self):
        # test opening a port, commands should be issued in order and the rule should be
        # recorded in 'rules' member of FirewallControl class instance
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.add_commands(
            CommandCaptureCommand(('/usr/sbin/lokkit', '-n', '-p', '%s:%s' % (self.port,
                                                                              self.proto))),
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.chain_output),
            CommandCaptureCommand(('/sbin/iptables', '-I', 'INPUT', '4', '-m', 'state', '--state',
                                   'new', '-p', self.proto, '--dport', self.port, '-j', 'ACCEPT')))

        response = self.test_firewall.add_rule(self.port, self.proto, self.desc, persist=True)

        # class instance should have record of added rule
        self.assertEqual(len(self.test_firewall.rules), 1)
        self.assertEqual(self.test_firewall.rules[0], self.example_port_rule)
        # None return value indicates success
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    @mock.patch.object(tempfile, 'mkstemp')
    def test_close_port(self, mock_mkstemp):
        # test closing a port, method includes file operations within a try clause which we
        # break from by raising an exception in the patched object. test only shell commands here
        e = IOError('abc')
        mock_mkstemp.side_effect = e

        self.add_commands(
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.chain_output),
            CommandCaptureCommand(('/sbin/iptables', '-D', 'INPUT', '-m', 'state', '--state',
                                   'new', '-p', self.proto, '--dport', self.port, '-j', 'ACCEPT')))

        try:
            self.test_firewall.remove_rule(self.port, self.proto, self.desc, persist=True)
            # this line should not be executed, we should receive an exception from the mock
            self.assertTrue(False)
        except IOError as e:
            self.assertTrue(e.args[0], 'abc')

        self.assertRanAllCommandsInOrder()

    @mock.patch.object(FirewallControlEL6, '_remove_port', return_value=FirewallControl.SuccessCode.UPDATED)
    def test_close_removes_rule(self, mock__remove_port):
        # test closing a port removes a rule from the class instance 'rules' list,
        # this is additionally required because test_close_port exits remove() before rule is
        # removed from list (intentionally thrown exception)

        # add rule object so we can test it is removed when we issue the remove() command
        self.test_firewall.rules.append(self.example_port_rule)
        self.assertEqual(len(self.test_firewall.rules), 1)

        # commands won't actually be issued because of the mock patch, example rule should be
        # removed from 'rules' list
        response = self.test_firewall.remove_rule(self.port, self.proto, self.desc, persist=True)

        # None return value indicates success
        self.assertEqual(response, None)
        # rule should have been removed from list
        self.assertEqual(len(self.test_firewall.rules), 0)

    def test_open_address(self):
        # test opening all ports on a specific address, commands should be issued in order and the
        # rule should be recorded in 'rules' member of FirewallControl class instance
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.add_commands(
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.chain_output),
            CommandCaptureCommand(('/sbin/iptables', '-I', 'INPUT', '4', '-m', 'state', '--state',
                                   'NEW', '-m', self.proto, '-p', self.proto, '-d', self.address,
                                   '-j', 'ACCEPT')))

        response = self.test_firewall.add_rule(0, self.proto, self.desc, persist=False, address=self.address)

        # class instance should have record of added rule
        self.assertEqual(len(self.test_firewall.rules), 1)
        self.assertEqual(self.test_firewall.rules[0], self.example_address_rule)
        # None return value indicates success
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    def test_close_address(self):
        # test remove_rule with address removes relevant rule from the class instance 'rules' list,
        # also test correct shell command is issued

        # add rule object so we can test it is removed when we issue the remove() command
        self.test_firewall.rules.append(self.example_address_rule)
        self.assertEqual(len(self.test_firewall.rules), 1)

        self.add_commands(
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.chain_output),
            CommandCaptureCommand(('/sbin/iptables', '-D', 'INPUT', '-m', 'state', '--state',
                                   'NEW', '-m', self.proto, '-p', self.proto, '-d', self.address,
                                   '-j', 'ACCEPT')))

        response = self.test_firewall.remove_rule(0, self.proto, self.desc, persist=False, address=self.address)

        # None return value indicates success
        self.assertEqual(response, None)
        # rule should have been removed from list
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.assertRanAllCommandsInOrder()

    def test_add_existing_port_rule(self):
        # rules are analysed to check they don't match existing entries, test that additional
        # rules are not added
        self.add_commands(
            CommandCaptureCommand(('/usr/sbin/lokkit', '-n', '-p', '%s:%s' % ('123', 'udp'))),
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.chain_output))

        response = self.test_firewall.add_rule('123', 'udp', 'test_service', persist=True)

        # None return value indicates success or no action performed
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    def test_add_existing_port_rule_different_index(self):
        # rules are analysed to check they don't match existing entries at different index,
        # test that additional rules are not added
        self.add_commands(
            CommandCaptureCommand(('/usr/sbin/lokkit', '-n', '-p', '%s:%s' % ('123', 'udp'))),
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.different_index_output))

        response = self.test_firewall.add_rule('123', 'udp', 'test_service', persist=True)

        # None return value indicates success or no action performed
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    def test_firewall_not_running(self):
        # if the firewall is not running, add_rule should silently exit while logging a warning
        # test that we return None in this situation
        self.add_commands(
            CommandCaptureCommand(('/usr/sbin/lokkit', '-n', '-p', '%s:%s' % ('123', 'udp'))),
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.not_running_msg, rc=3))

        response = self.test_firewall.add_rule('123', 'udp', 'test_service', persist=True)

        # None return value indicates success or no action performed
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    def test_firewall_not_configured(self):
        # if the firewall is not configured, add_rule should silently exit while logging a warning
        # test that we return None in this situation
        self.add_commands(
            CommandCaptureCommand(('/usr/sbin/lokkit', '-n', '-p', '%s:%s' % ('123', 'udp'))),
            CommandCaptureCommand(('service', 'iptables', 'status'), stdout=self.not_configured_output, rc=3))

        response = self.test_firewall.add_rule('123', 'udp', 'test_service', persist=True)

        # None return value indicates success or no action performed
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()


class TestFirewallControlEL7(BaseTestFC.BaseTestFirewallControl):

    not_running_msg = 'FirewallD is not running'

    def __init__(self, *args, **kwargs):
        super(TestFirewallControlEL7, self).__init__(*args, **kwargs)
        self.init_firewall('7.2')

    def test_open_port(self):
        # test opening a port, commands should be issued in order and the rule should be
        # recorded in 'rules' member of FirewallControl class instance
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.add_commands(
            CommandCaptureCommand(('/usr/bin/firewall-cmd',
                                   '--add-port=%s/%s' % (self.port, self.proto))),
            CommandCaptureCommand(('/usr/bin/firewall-cmd',
                                   '--add-port=%s/%s' % (self.port, self.proto),
                                   '--permanent')))

        response = self.test_firewall.add_rule(self.port, self.proto, self.desc, persist=True)

        # class instance should have record of added rule
        self.assertEqual(len(self.test_firewall.rules), 1)
        self.assertEqual(self.test_firewall.rules[0], self.example_port_rule)
        # None return value indicates success
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    def test_close_port(self):
        # test closing a port, commands should be issued in order and the rule should be
        # removed from 'rules' member of FirewallControl class instance

        # add rule object so we can test it is removed when we issue the remove() command
        self.test_firewall.rules.append(self.example_port_rule)
        self.assertEqual(len(self.test_firewall.rules), 1)

        self.add_commands(
            CommandCaptureCommand(('/usr/bin/firewall-cmd',
                                   '--remove-port=%s/%s' % (self.port, self.proto))),
            CommandCaptureCommand(('/usr/bin/firewall-cmd',
                                   '--remove-port=%s/%s' % (self.port, self.proto),
                                   '--permanent')))

        response = self.test_firewall.remove_rule(self.port, self.proto, self.desc, persist=True)

        # None return value indicates success
        self.assertEqual(response, None)
        # rule should have been removed from list
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.assertRanAllCommandsInOrder()

    def test_open_address(self):
        # test opening all ports on a specific address, commands should be issued in order and the
        # rule should be recorded in 'rules' member of FirewallControl class instance
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.add_commands(
            CommandCaptureCommand(('/usr/bin/firewall-cmd', '--add-rich-rule='
                                                            'rule family="ipv4" '
                                                            'destination address="%s" '
                                                            'protocol value="%s" '
                                                            'accept' % (self.address, self.proto))))

        response = self.test_firewall.add_rule(0, self.proto, self.desc, persist=False,
                                               address=self.address)

        # None return value indicates success
        self.assertEqual(response, None)
        # class instance should have record of added rule
        self.assertEqual(len(self.test_firewall.rules), 1)
        self.assertEqual(self.test_firewall.rules[0], self.example_address_rule)
        self.assertRanAllCommandsInOrder()

    def test_close_address(self):
        # test remove_rule with address removes relevant rule from the class instance 'rules' list,
        # also test correct shell command is issued

        # add rule object so we can test it is removed when we issue the remove() command
        self.test_firewall.rules.append(self.example_address_rule)
        self.assertEqual(len(self.test_firewall.rules), 1)

        self.add_commands(
            CommandCaptureCommand(('/usr/bin/firewall-cmd', '--remove-rich-rule='
                                                            'rule family="ipv4" '
                                                            'destination address="%s" '
                                                            'protocol value="%s" '
                                                            'accept' % (self.address, self.proto))))

        response = self.test_firewall.remove_rule(0, self.proto, self.desc, persist=False, address=self.address)

        # None return value indicates success
        self.assertEqual(response, None)
        # rule should have been removed from list
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.assertRanAllCommandsInOrder()

    def test_firewall_not_running(self):
        # if the firewall is not running, add_rule should silently exit while logging a warning
        # test that we return None in this situation
        self.assertEqual(len(self.test_firewall.rules), 0)
        self.add_command(('/usr/bin/firewall-cmd', '--add-port=%s/%s' % (self.port, self.proto)),
                         rc=252, stdout=self.not_running_msg)

        response = self.test_firewall.add_rule(self.port, self.proto, self.desc, persist=True)

        # class instance should have record of added rule
        self.assertEqual(len(self.test_firewall.rules), 0)
        # None return value indicates success
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()
